"""Fix some compatibility issues between Python versions."""

import collections
import typing as ty

if not hasattr(collections, "Callable"):
    collections.Callable = ty.Callable


def enable_compat():
    """No-op compat."""



"""Json source."""

import typing as ty
import warnings
from json import JSONDecodeError
from pathlib import Path

from koyo.json import read_json_data, write_json_data
from koyo.typing import PathLike


class JSONCache:
    """Cache that data is stored in a JSON file."""

    FILENAME: str

    def __init__(self, path: PathLike):
        self._dir_path = Path(path)

    def __getattr__(self, item: str):
        return self[item]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}<{self.as_str()}>"

    def __getitem__(self, key: str) -> ty.Any:
        return self.read_key(key)

    def __setitem__(self, key: str, value: bool):
        self.write_key(key, value)

    def __contains__(self, item):
        return item in self.read()

    def __delitem__(self, key: str) -> None:
        data = self.read()
        if key in data:
            del data[key]
            self.write(data)

    def remove(self, key: str) -> None:
        """Remove key from cache."""
        data = self.read()
        if key in data:
            del data[key]
            self.write(data)

    def remove_multiple(self, *keys: str) -> None:
        """Remove multiple keys from cache."""
        data = self.read()
        for key in keys:
            if key in data:
                del data[key]
        self.write(data)

    def _set_default(self):
        data = {}
        self.write(data)

    def clear(self):
        """Remove existing tags."""
        self.write({})

    @property
    def name(self) -> str:
        """Return basename of the directory."""
        return self._dir_path.name

    def keys(self):
        """Return all keys in cache."""
        return self.read().keys()

    def items(self) -> ty.ItemsView:
        """Return all items in cache."""
        return self.read().items()

    def update(self, **kwargs: ty.Any) -> None:
        """Update cache with new key-value pairs."""
        data = self.read()
        data.update(**kwargs)
        self.write(data)

    def get(self, key: str, default=None):
        """Get key."""
        return self.get_key(key, default)

    @property
    def path(self) -> Path:
        """Get path of flag's data."""
        return self._dir_path / self.FILENAME

    def exists(self) -> bool:
        """Check whether flags file exists."""
        return self.path.exists()

    def as_str(self, sep="; ", exclude: ty.Optional[ty.Tuple[str, ...]] = None) -> str:
        """Get string representation of the flags."""
        if exclude is None:
            exclude = ()
        if self.exists():
            data = self.read()
            ret = ""
            n_stop = len(data) - 1
            for i, (k, v) in enumerate(data.items()):
                if k in exclude:
                    continue
                ret += f"{k}: {v}"
                if i != n_stop:
                    ret += sep
            return ret
        return ""

    def as_dict(self, exclude: ty.Optional[ty.Tuple[str, ...]] = None) -> dict:
        """Get dictionary representation of the flags."""
        if exclude is None:
            exclude = ()
        if self.exists():
            data = self.read()
            return {k: v for k, v in data.items() if k not in exclude}
        return {}

    def print_summary(self, name: str = "", pre: str = "\t", sep: str = "\n"):
        """Print summary about JSON store."""
        if name:
            print(name)
        data = self.read()
        if not data:
            print(f"{pre}<no data>")
        else:
            for k, v in data.items():
                print(f"{pre}{k}: {v}", sep=sep)

    @staticmethod
    def format_value(v) -> ty.Any:
        """Format value."""
        if isinstance(v, str):
            if v.lower() == "true":
                return True
            elif v.lower() == "false":
                return False
            elif v.isnumeric():
                return int(v)
            try:
                return float(v)
            except ValueError:
                return v
        return v

    def read(self) -> ty.Dict:
        """Read data."""
        if self.exists():
            try:
                data = read_json_data(self.path)
                data = {k: self.format_value(v) for k, v in data.items()}
                return data
            except JSONDecodeError:
                warnings.warn(f"Failed to read JSON file: {self.path}")
        return {}

    def write(self, data: ty.Dict):
        """Write data to disk."""
        self._dir_path.mkdir(parents=True, exist_ok=True)
        write_json_data(self.path, data)

    def save(self, path: PathLike):
        """Save data to disk."""
        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)
        write_json_data(path, self.read())

    def get_key(self, key: str, default=None):
        """Read flag from the flag file."""
        if not self.exists():
            self._set_default()
        return self.read().get(key, default)

    def read_key(self, key: str):
        """Read flag from the flag file."""
        if not self.exists():
            self._set_default()
        return self.read()[key]

    def write_key(self, key: str, value: ty.Union[int, float, str, bool]):
        """Write flag to the flag file."""
        if not self.exists():
            self._set_default()
        data = self.read()
        data[key] = value
        self.write(data)

    def get_for_keys(self, keys: list[str], default: ty.Any = None) -> dict[str, ty.Any]:
        """Get values for specified keys."""
        return {k: self.get_key(k, default) for k in keys}



"""Logging utilities."""

from __future__ import annotations

import os
import sys
import typing as ty
from functools import partial
from pathlib import Path

from koyo.typing import PathLike
from koyo.utilities import find_nearest_value

if ty.TYPE_CHECKING:
    from loguru import Logger
LOGGER_TO_PATH: dict[PathLike, int] = {}

LEVEL_FORMAT = "<level>{level: <8}</level>"
TIME_FORMAT = "{time:YYYY-MM-DD HH:mm:ss:SSS}"
LOG_FMT = "[LEVEL_FORMAT][TIME_FORMAT][{process}] {message}".replace("TIME_FORMAT", TIME_FORMAT).replace(
    "LEVEL_FORMAT", LEVEL_FORMAT
)
COLOR_LOG_FMT = "<green>[LEVEL_FORMAT]</green><cyan>[TIME_FORMAT]</cyan><red>[{process}]</red> {message}".replace(
    "TIME_FORMAT", TIME_FORMAT
).replace("LEVEL_FORMAT", LEVEL_FORMAT)


def timed_call(log_func: ty.Callable, message: str, *args: ty.Any, **kwargs: ty.Any) -> ty.Any:
    """Time a function call."""
    from koyo.timer import MeasureTimer

    with MeasureTimer() as timer:
        yield
    log_func(message + f" in {timer()}", *args, **kwargs)


def get_stdout() -> ty.TextIO:
    """Get stdout."""
    if sys.stdout is None:
        sys.stdout = sys.stderr
    if sys.stdout is None:
        sys.stdout = open(os.devnull, "w")
    return sys.stdout


def get_stderr() -> ty.TextIO:
    """Get stderr."""
    if sys.stderr is None:
        sys.stderr = sys.stdout
    if sys.stderr is None:
        sys.stderr = open(os.devnull, "w")
    return sys.stderr


def set_logger(
    verbosity: int,
    no_color: bool,
    log: PathLike | None = None,
    logger: Logger | None = None,
    enable: tuple[str, ...] = ("koyo",),
):
    """Setup logger."""
    if logger is None:
        from loguru import logger

    level = verbosity * 10
    level, fmt, colorize, enqueue = get_loguru_config(level, no_color=no_color)
    set_loguru_env(fmt, level, colorize, enqueue)
    set_loguru_log(level=level.upper(), no_color=no_color, logger=logger)
    for enable_ in enable:
        logger.enable(enable_)
    set_loguru_log(level=level.upper(), no_color=no_color)
    # override koyo logger
    if log:
        set_loguru_log(
            log,
            level=level.upper(),  # type: ignore[attr-defined]
            enqueue=enqueue,
            colorize=False,
            no_color=True,
            catch=True,
            diagnose=True,
            logger=logger,
            remove=False,
        )
    logger.debug(f"Activated logger with level '{level}'.")
    logger.trace(f"Command: {' '.join(sys.argv)}")


def get_logger() -> Logger:
    """Get logger with extra functions."""
    from loguru import logger

    logger.timed_trace = partial(timed_call, logger.trace)
    logger.timed_debug = partial(timed_call, logger.debug)
    logger.timed_info = partial(timed_call, logger.info)
    logger.timed_success = partial(timed_call, logger.success)
    logger.timed_warning = partial(timed_call, logger.warning)
    logger.timed_error = partial(timed_call, logger.error)
    logger.timed_critical = partial(timed_call, logger.critical)
    logger.timed_exception = partial(timed_call, logger.exception)
    return logger


def set_loguru_env(fmt: str, level: str, enqueue: bool, colorize: bool) -> None:
    """Set loguru environment variables."""
    os.environ["LOGURU_AUTOINIT"] = "0"
    os.environ["LOGURU_FORMAT"] = str(fmt)
    os.environ["LOGURU_LEVEL"] = str(level)
    os.environ["LOGURU_ENQEUE"] = str(enqueue)
    os.environ["LOGURU_COLORIZE"] = str(colorize)


def get_loguru_level(level: int | str) -> str:
    """Get loguru level."""
    if isinstance(level, str):
        return level.upper()
    levels = {
        0: "trace",
        5: "trace",
        10: "debug",
        20: "info",
        25: "success",
        30: "warning",
        40: "error",
        50: "critical",
    }
    if level not in levels:
        level = find_nearest_value(list(levels.keys()), level)
    level = levels[level]
    return level.upper()


def get_loguru_env() -> tuple[str, str, bool, bool]:
    """Get logoru environment variables."""
    colorize = bool(os.environ.get("LOGURU_COLORIZE", "True"))
    level = os.environ.get("LOGURU_LEVEL", "info")
    enqueue = bool(os.environ.get("LOGURU_ENQEUE", "True"))
    fmt = os.environ.get("LOGURU_FORMAT", LOG_FMT if not colorize else COLOR_LOG_FMT)
    return fmt, level, enqueue, colorize


def get_loguru_config(
    level: str | int | float,
    no_color: bool,
    enqueue: bool = True,
    fmt: str = LOG_FMT,
    color_fmt: str = COLOR_LOG_FMT,
) -> tuple[str, str, bool, bool]:
    """Return level."""
    level = get_loguru_level(level)
    colorize = not no_color
    fmt = fmt if no_color else color_fmt
    return level.upper(), fmt, colorize, enqueue


def set_loguru_log(
    sink=sys.stderr,
    level: str | int = 20,
    no_color: bool = False,
    enqueue: bool = True,
    fmt: str | None = None,
    diagnose: bool = False,
    catch: bool = False,
    colorize: bool | None = None,
    remove: bool = True,
    logger: Logger | None = None,
) -> int:
    """Set loguru formatting."""
    if logger is None:
        from loguru import logger
    if logger is None:
        raise ValueError("Logger is None - cannot set loguru log.")
    if sink is None:
        sink = get_stderr()
    if sink is None:
        raise ValueError("Sink is None - cannot set loguru log.")
    if colorize is not None:
        no_color = not colorize

    # automatically get format
    fmt = fmt if fmt is not None else (LOG_FMT if no_color else COLOR_LOG_FMT)
    level = get_loguru_level(level)

    if remove:
        logger.remove(None)
    return logger.add(
        sink, level=level, format=fmt, colorize=not no_color, enqueue=enqueue, diagnose=diagnose, catch=catch
    )


class LoggerMixin:
    """Logger mixin."""

    dataset_name: str

    # LoggerMixin
    log_dir: Path
    LOG_MODULE: str
    LOG_PATH: Path | None = None

    def is_logging_enabled(self) -> bool:
        """Check if logging is enabled."""
        raise NotImplementedError("Must implement method")

    def set_log(self, log: bool | None = None, remove: bool = False) -> None:
        """Set logging."""
        from loguru import logger

        if log is None:
            log = self.is_logging_enabled()

        global LOGGER_TO_PATH

        logger.enable(self.LOG_MODULE)
        if not log:
            return
        if remove:
            logger.remove()

        _, level, enqueue, colorize = get_loguru_env()
        # add stderr logger
        if "stdout" not in LOGGER_TO_PATH:
            LOGGER_TO_PATH["stdout"] = set_loguru_log(
                level=level,
                enqueue=enqueue,
                colorize=colorize,
                catch=True,
                diagnose=True,
                logger=logger,
                remove=False,
            )
            logger.info("Setup logging to file - 'stderr'")
        # add file logger
        if not self.LOG_PATH:
            filename = "log"
            if self.dataset_name:
                filename += f"_dataset={self.dataset_name}"
            else:
                filename += "_main"
            self.LOG_PATH = (self.log_dir / filename).with_suffix(".txt")
            if self.LOG_PATH not in LOGGER_TO_PATH:
                LOGGER_TO_PATH[self.LOG_PATH] = set_loguru_log(
                    self.LOG_PATH,
                    level=level,
                    enqueue=enqueue,
                    colorize=False,
                    no_color=True,
                    catch=True,
                    diagnose=True,
                    logger=logger,
                    remove=False,
                )
                logger.info(f"Setup logging to file - '{self.LOG_PATH!s}'")



import inspect
import sys


def open_file_audit_hook(name, *args):
    """Open file hook."""
    if name == "open":
        print(name, *args, "was called:")
        caller = inspect.currentframe()
        while caller := caller.f_back:
            print(f"\tFunction {caller.f_code.co_name} " f"in {caller.f_code.co_filename}:" f"{caller.f_lineno}")


def install_open_file_audit_hook():
    """Add open file audit hook."""
    sys.addaudithook(open_file_audit_hook)



"""System utilities."""

from __future__ import annotations

import contextlib
import inspect
import os
import platform
import subprocess
import sys
from functools import lru_cache




IS_WIN = sys.platform == "win32"
IS_LINUX = sys.platform == "linux"
IS_MAC = sys.platform == "darwin"
IS_MAC_ARM = IS_MAC and platform.processor() == "arm"
IS_PYINSTALLER = getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")


def is_envvar(key: str, value: str) -> bool:
    """Check if an environment variable is set."""
    return key in os.environ and os.environ[key] == str(value)


def is_envvar_set(key: str) -> bool:
    """Check if an environment variable is set."""
    return key in os.environ and os.environ[key]


def set_freeze_support() -> None:
    """Set freeze support for multiprocessing."""
    import sys
    from multiprocessing import freeze_support, set_start_method

    freeze_support()
    if sys.platform == "darwin":
        set_start_method("spawn", True)


def get_cli_path(name: str, env_key: str = "", default: str = "") -> str:
    """Get path to imimspy executable.

    The path is determined in the following order:
    1. First, we check whether environment variable `{env_key}_{name.upper()}_PATH` is set.
    2. If not, we check whether we are running as a PyInstaller app.
    3. If not, we check whether we are running as a Python app.
    4. If not, we raise an error.
    """
    import os
    import sys
    from pathlib import Path

    from koyo.utilities import running_as_pyinstaller_app

    env_var = f"{env_key}_{name.upper()}_PATH"
    if os.environ.get(env_var, None):
        script_path = Path(os.environ[env_var])
        if script_path.exists():
            return str(script_path)

    base_path = Path(sys.executable).parent
    if running_as_pyinstaller_app():
        if IS_WIN:
            script_path = base_path / f"{name}.exe"
        elif IS_MAC or IS_LINUX:
            script_path = base_path / name
        else:
            raise NotImplementedError(f"Unsupported OS: {sys.platform}")
        if script_path.exists():
            return str(script_path)
    else:
        # on Windows, {name} lives under the `Scripts` directory
        if IS_WIN:
            script_path = base_path
            if script_path.name != "Scripts":
                script_path = base_path / "Scripts"
            if script_path.exists() and (script_path / f"{name}.exe").exists():
                return str(script_path / f"{name}.exe")
        elif IS_MAC or IS_LINUX:
            script_path = base_path / name
            if script_path.exists():
                return str(script_path)
        else:
            script_path = base_path / f"{name}.exe"
            if script_path.exists():
                return str(script_path)
    if default:
        return default
    raise RuntimeError(f"Could not find '{name}' executable.")


def who_called_me() -> tuple[str, str, int]:
    """Get the file name, function name, and line number of the caller."""
    # Get the current frame
    current_frame = inspect.currentframe()
    # Get the caller's frame
    caller_frame = current_frame.f_back

    # Extract file name, line number, and function name
    file_name = caller_frame.f_code.co_filename
    line_number = caller_frame.f_lineno
    function_name = caller_frame.f_code.co_name
    print(f"Called from file: {file_name}, function: {function_name}, line: {line_number}")
    return file_name, function_name, line_number


def who_called_me_stack(n: int = 6) -> None:
    """Print the last 5 callers that led to this function being called."""
    stack = inspect.stack()
    # The stack list starts with the current frame at index 0,
    # so the callers start from index 1 onward.
    # We will retrieve up to 5 callers, or fewer if the stack isn't that deep.
    limit = min(n, len(stack))  # 1 (immediate caller) + up to 5 total callers
    for i in range(1, limit):
        frame_info = stack[i]
        file_name = frame_info.filename
        function_name = frame_info.function
        line_number = frame_info.lineno
        print(f"Caller #{i}: File '{file_name}', Function '{function_name}', Line {line_number}")


def _linux_sys_name() -> str:
    """
    Try to discover linux system name base on /etc/os-release file or lsb_release command output
    https://www.freedesktop.org/software/systemd/man/os-release.html.
    """
    os_release_path = "/etc/os-release"

    if os.path.exists(os_release_path):
        with open(os_release_path) as f_p:
            data = {}
            for line in f_p:
                field, value = line.split("=")
                data[field.strip()] = value.strip().strip('"')
        if "PRETTY_NAME" in data:
            return data["PRETTY_NAME"]
        if "NAME" in data:
            if "VERSION" in data:
                return f"{data['NAME']} {data['VERSION']}"
            if "VERSION_ID" in data:
                return f"{data['NAME']} {data['VERSION_ID']}"
            return f"{data['NAME']} (no version)"
    return _linux_sys_name_lsb_release()


def _linux_sys_name_lsb_release() -> str:
    """Try to discover linux system name base on lsb_release command output."""
    with contextlib.suppress(subprocess.CalledProcessError):
        res = subprocess.run(["lsb_release", "-d", "-r"], check=True, capture_output=True)
        text = res.stdout.decode()
        data = {}
        for line in text.split("\n"):
            key, val = line.split(":")
            data[key.strip()] = val.strip()
        version_str = data["Description"]
        if not version_str.endswith(data["Release"]):
            version_str += " " + data["Release"]
        return version_str
    return ""


def _sys_name() -> str:
    """Discover MacOS or Linux Human readable information. For Linux provide information about distribution."""
    with contextlib.suppress(Exception):
        if sys.platform == "linux":
            return _linux_sys_name()
        if sys.platform == "darwin":
            with contextlib.suppress(subprocess.CalledProcessError):
                res = subprocess.run(
                    ["sw_vers", "-productVersion"],
                    check=True,
                    capture_output=True,
                )
                return f"MacOS {res.stdout.decode().strip()}"
    return ""


def get_module_path(module: str, filename: str) -> str:
    """Get module path."""
    import importlib.resources

    if not filename.endswith(".py"):
        filename += ".py"

    path = str(importlib.resources.files(module).joinpath(filename))
    return path


def running_as_pyinstaller_app() -> bool:
    """Infer whether we are running pyinstaller bundle."""
    import sys

    return getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")


@lru_cache(maxsize=4)
def running_as_briefcase_app() -> bool:
    """Infer whether we are running as a briefcase bundle."""
    import importlib_metadata

    # https://github.com/beeware/briefcase/issues/412
    # https://github.com/beeware/briefcase/pull/425
    # note that a module may not have a __package__ attribute
    try:
        app_module = sys.modules["__main__"].__package__
    except AttributeError:
        return False
    try:
        metadata = importlib_metadata.metadata(app_module)
    except importlib_metadata.PackageNotFoundError:
        return False
    return "Briefcase-Version" in metadata





"""Image processing functions."""

import typing as ty

import numpy as np


def clip_hotspots(img: np.ndarray, quantile: float = 0.99) -> np.ndarray:
    """Remove hotspots from ion image.

    Parameters
    ----------
    img : np.ndarray
        Image array.
    quantile : float
        Quantile at which to clip intensities.

    Returns
    -------
    img : np.ndarray
        Clipped array.

    """
    mask = np.isnan(img)
    img = np.nan_to_num(img, posinf=0, neginf=0)
    min_visible = np.max(img) / 256
    if min_visible > 0:
        hotspot_threshold = np.quantile(img[img > min_visible], quantile)
        img = np.clip(img, None, hotspot_threshold)
    img[mask] = np.nan
    return img


def reshape_array(array: np.ndarray, image_shape: ty.Tuple[int, int], pixel_index: np.ndarray, fill_value: float = 0):
    """
    Reshape 1D data to 2D heatmap.

    Parameters
    ----------
    array: np.array / list
        1D array of values to be reshaped
    image_shape: tuple
        final shape of the image
    pixel_index: np.array
        array containing positions where pixels should be placed, considering missing values -
        e.g. not acquired pixels
    fill_value : float, optional
        if value is provided, it will be used to fill-in the values

    Returns
    -------
    im_array: np.array
        reshaped heatmap of shape `image_shape`
    """
    if isinstance(array, np.matrix):
        array = np.asarray(array).flatten()
    array = np.asarray(array)
    dtype = np.float32 if isinstance(fill_value, float) else array.dtype

    image_n_pixels = np.prod(image_shape)
    im_array = np.full(image_n_pixels, dtype=dtype, fill_value=fill_value)
    im_array[pixel_index] = array
    im_array = np.reshape(im_array, image_shape)
    return im_array


def unshape_array(image: np.ndarray, pixel_index: np.ndarray) -> np.ndarray:
    """Retrieve original vector of intensities from an image."""
    if image.ndim == 3:  # channels, height, width
        return image.reshape(image.shape[0], -1)[:, pixel_index].T
    return image.reshape(-1)[pixel_index]


def reshape_array_from_coordinates(
    array: np.ndarray,
    image_shape: ty.Tuple[int, int],
    coordinates: np.ndarray,
    fill_value: float = 0,
    offset: int = 0,
):
    """Reshape array based on xy coordinates.

    This function assumes that the coordinates are 1-indexed.
    """
    dtype = np.float32 if np.isnan(fill_value) else array.dtype
    im = np.full(image_shape, fill_value=fill_value, dtype=dtype)
    try:
        im[coordinates[:, 1] - offset, coordinates[:, 0] - offset] = array
    except IndexError:
        im[coordinates[:, 0] - offset, coordinates[:, 1] - offset] = array
    return im


def flatten_array_from_coordinates(array: np.ndarray, coordinates: np.ndarray, offset: int = 0) -> np.ndarray:
    """Flatten array based on xy coordinates."""
    if array.ndim == 2:
        try:
            return array[coordinates[:, 1] - offset, coordinates[:, 0] - offset]
        except IndexError:
            return array[coordinates[:, 0] - offset, coordinates[:, 1] - offset]
    else:
        try:
            res = array[:, coordinates[:, 1] - offset, coordinates[:, 0] - offset]
        except IndexError:
            res = array[:, coordinates[:, 0] - offset, coordinates[:, 1] - offset]
        # need to swap axes
        res = np.swapaxes(res, 0, 1)
        return res


def reshape_array_batch(
    array: np.ndarray, image_shape: ty.Tuple[int, int], pixel_index: np.ndarray, fill_value: float = 0
):
    """Reshape many images into a data cube."""
    array = np.asarray(array)
    if array.ndim == 1:
        return reshape_array(array, image_shape, pixel_index, fill_value)
    count = array.shape[1]
    dtype = np.float32 if isinstance(fill_value, float) else array.dtype

    im_array = np.full((count, np.prod(image_shape)), dtype=dtype, fill_value=fill_value)
    for i in range(count):
        im_array[i, pixel_index] = array[:, i]
    # reshape data
    im_array = np.reshape(im_array, (count, *image_shape))
    return im_array


def reshape_array_batch_from_coordinates(
    array: np.ndarray,
    image_shape: ty.Tuple[int, int],
    coordinates: np.ndarray,
    fill_value: int = 0,
    offset: int = 0,
):
    """Batch reshape image.

    This function assumes that the coordinates are 1-indexed.
    """
    if array.ndim != 2:
        raise ValueError("Expected 2-D array.")
    n_features = array.shape[1]
    dtype = np.float32 if np.isnan(fill_value) else array.dtype
    im = np.full((n_features, *image_shape), fill_value=fill_value, dtype=dtype)
    try:
        for i in range(n_features):
            im[i, coordinates[:, 1] - offset, coordinates[:, 0] - offset] = array[:, i]
    except IndexError:
        for i in range(n_features):
            im[i, coordinates[:, 0] - offset, coordinates[:, 1] - offset] = array[:, i]
    return im


def get_coordinates_from_index(index: np.ndarray, shape: ty.Tuple[int, int]) -> np.ndarray:
    """Convert frame index to xy coordinates."""
    index = np.asarray(index)
    # generate image shape
    index_im = reshape_array(index, shape, index)
    if shape[0] != index_im.shape[0]:
        raise ValueError("Image dimension 0 does not match that of the dataset")
    if shape[1] != index_im.shape[1]:
        raise ValueError("Image dimension 1 does not match that of the dataset")

    y_coordinates, x_coordinates = np.indices(index_im.shape)
    yx_coordinates = np.c_[np.ravel(y_coordinates), np.ravel(x_coordinates)][index]
    return yx_coordinates


def colocalization(img_a: np.ndarray, img_b: np.ndarray) -> float:
    """Calculate degree of colocalization between two ion images.

    This implementation is nearly identical to that of METASPACE.

    Citation: Ovchinnikova et al. (2020) ColocML. https://doi.org/10.1093/bioinformatics/btaa085

    Parameters
    ----------
    img_a: np.ndarray
        first image
    img_b : np.ndarray
        second image

    Returns
    -------
    similarity : float
        similarity score between two images
    """
    from scipy.ndimage import median_filter
    from sklearn.metrics.pairwise import cosine_similarity

    img_a = np.nan_to_num(img_a)
    img_b = np.nan_to_num(img_b)
    h, w = img_a.shape

    def _preprocess(img):
        img = img.copy().reshape((h, w))
        img[img < np.quantile(img, 0.5)] = 0
        return median_filter(img, (3, 3)).reshape([1, h * w])

    return cosine_similarity(_preprocess(img_a), _preprocess(img_b))[0, 0]


def colocalization_matrix(images: np.ndarray, size: ty.Tuple[int, ...] = (3, 3)) -> np.ndarray:
    """
    Calculates level of colocalization between all pairs of images in a list of ion images.
    If many checks are needed, it is usually faster to generate the entire matrix than to do
    many separate calls to "colocalization".

    Citation: Ovchinnikova et al. (2020) ColocML. https://doi.org/10.1093/bioinformatics/btaa085

    Parameters
    ----------
    images : List[np.ndarray]
        list of ion images
    size : Tuple[int, ...]
        size of the median filter.

    Returns
    -------
    similarity_matrix : np.ndarray
        colocalization similarity matrix
    """
    from scipy.ndimage import median_filter
    from sklearn.metrics.pairwise import pairwise_kernels

    assert images.ndim == 3, "Expected 2D array."  # flat images

    count = images.shape[0]
    if count == 0:
        similarity_matrix = np.ones((0, 0))
    elif count == 1:
        similarity_matrix = np.ones((1, 1))
    else:
        h, w = images[0].shape
        flat_images = np.vstack([i.flatten() for i in images])
        flat_images[flat_images < np.quantile(flat_images, 0.5, axis=1, keepdims=True)] = 0
        filtered_images = median_filter(images.reshape((count, h, w)), (1, *size)).reshape((count, h * w))
        similarity_matrix = pairwise_kernels(filtered_images, metric="cosine")
    return similarity_matrix


def pearson_similarity(img_a: np.ndarray, img_b: np.ndarray, size: ty.Tuple[int, int] = (3, 3)) -> float:
    """Calculate degree of similarity between two images using Pearson correlation.

    Parameters
    ----------
    img_a : np.ndarray
        First image array.
    img_b : np.ndarray
        Second image array.
    size : tuple
        Size of the median filter.

    Returns
    -------
    score : float
        Result of linear regression after each image had median filter applied to it.

    """
    from scipy.ndimage import median_filter
    from scipy.stats import linregress

    if len(size) != 2:
        raise ValueError("Median filter expected 2-element tuple.")

    img_a = median_filter(img_a, size)
    img_b = median_filter(img_b, size)
    mask = (img_a > 0) & (img_b > 0)
    return linregress(img_a[mask], img_b[mask]).rvalue



"""Affine transformation functions."""

from __future__ import annotations

import numpy as np


def transform_xy_coordinates(
    xy: np.ndarray, *, yx_affine: np.ndarray | None = None, xy_affine: np.ndarray | None = None
) -> np.ndarray:
    """Transform xy coordinates using either yx or xy affine matrix."""
    if xy_affine is None and yx_affine is None:
        raise ValueError("Either xy_affine or yx_affine should be provided.")
    if xy_affine is not None and yx_affine is not None:
        raise ValueError("Only one of xy_affine or yx_affine should be provided.")
    xy = np.hstack([xy, np.ones((xy.shape[0], 1))])
    if yx_affine is not None:
        xy = np.dot(xy, yx_affine.T)
    if xy_affine is not None:
        xy = np.dot(xy, xy_affine)
    xy = xy[:, :2]
    return xy



import numpy as np
import scipy.sparse as sparse


def get_index_dtype(arrays=(), maxval=None, check_contents=False):
    """
    Based on input (integer) arrays `a`, determine a suitable index data
    type that can hold the data in the arrays.

    Parameters
    ----------
    arrays : tuple of array_like
        Input arrays whose types/contents to check
    maxval : float, optional
        Maximum value needed
    check_contents : bool, optional
        Whether to check the values in the arrays and not just their types.
        Default: False (check only the types)

    Returns
    -------
    dtype : dtype
        Suitable index data type (int32 or int64)

    """
    int32min = np.iinfo(np.int32).min
    int32max = np.iinfo(np.int32).max

    dtype = np.intc
    if maxval is not None:
        if maxval > int32max:
            dtype = np.int64

    if isinstance(arrays, np.ndarray):
        arrays = (arrays,)

    for arr in arrays:
        arr = np.asarray(arr)
        if not np.can_cast(arr.dtype, np.int32):
            if check_contents:
                if arr.size == 0:
                    # a bigger type not needed
                    continue
                elif np.issubdtype(arr.dtype, np.integer):
                    maxval = arr.max()
                    minval = arr.min()
                    if minval >= int32min and maxval <= int32max:
                        # a bigger type not needed
                        continue

            dtype = np.int64
            break

    return dtype


def asindices(x):
    try:
        x = np.asarray(x)

        # Check index contents to avoid creating 64bit arrays needlessly
        idx_dtype = get_index_dtype((x,), check_contents=True)
        if idx_dtype != x.dtype:
            x = x.astype(idx_dtype)
    except Exception:
        raise IndexError("invalid index")
    else:
        return x


def check_bounds(indices, N):
    if indices.size == 0:
        return (0, 0)

    max_indx = indices.max()
    if max_indx >= N:
        raise IndexError("index (%d) out of range" % max_indx)

    min_indx = indices.min()
    if min_indx < -N:
        raise IndexError("index (%d) out of range" % (N + min_indx))

    return min_indx, max_indx


def check_ellipsis(index):
    """Process indices with Ellipsis. Returns modified index."""
    if index is Ellipsis:
        return (slice(None), slice(None))
    elif isinstance(index, tuple):
        # Find first ellipsis
        for j, v in enumerate(index):
            if v is Ellipsis:
                first_ellipsis = j
                break
        else:
            first_ellipsis = None

        # Expand the first one
        if first_ellipsis is not None:
            # Shortcuts
            if len(index) == 1:
                return (slice(None), slice(None))
            elif len(index) == 2:
                if first_ellipsis == 0:
                    if index[1] is Ellipsis:
                        return (slice(None), slice(None))
                    else:
                        return (slice(None), index[1])
                else:
                    return (index[0], slice(None))

            # General case
            tail = ()
            for v in index[first_ellipsis + 1 :]:
                if v is not Ellipsis:
                    tail = (*tail, v)
            nd = first_ellipsis + len(tail)
            nslice = max(0, 2 - nd)
            return index[:first_ellipsis] + (slice(None),) * nslice + tail
    return index


def unpack_index(index):
    index = check_ellipsis(index)

    if isinstance(index, tuple):
        if len(index) == 2:
            row, col = index
        elif len(index) == 1:
            row, col = index[0], slice(None)
    else:
        row, col = index, slice(None)

    return row, col


def extractor(indices, N, fmt="csr"):
    """
    row = np.arange(500)
    P = extractor(row, shape[0]).
    """
    indices = asindices(indices).copy()

    min_indx, __ = check_bounds(indices, N)

    if min_indx < 0:
        indices[indices < 0] += N

    indptr = np.arange(len(indices) + 1, dtype=indices.dtype)
    data = np.ones(len(indices), dtype=np.int32)
    shape = (len(indices), N)
    return sparse.csr_matrix((data, indices, indptr), shape=shape, dtype=np.int32, copy=False).asformat(fmt)


def get_array_statistics(array) -> str:
    """Calculate array statistics."""
    from koyo.utilities import calculate_array_size

    statistics: str = calculate_array_sparsity(array)
    size: str = calculate_array_size(array)
    dense_size: str = estimate_dense_size(array)

    return statistics + " | " + size + " | [As dense: " + dense_size + "]"


def estimate_dense_size(array) -> str:
    """Calculate approximate size of the array if it was dense."""
    from koyo.utilities import format_size

    total_val = float(np.prod(array.shape))
    n_bytes = total_val * array.dtype.itemsize
    return format_size(n_bytes)


def calculate_array_sparsity(array):
    """Calculate the sparsity and density of an array."""
    if hasattr(array, "nnz"):
        non_zero = array.nnz
        fmt = array.format
    else:
        non_zero = np.count_nonzero(array)
        fmt = "dense"
    total_val = np.prod(array.shape)
    sparsity = (total_val - non_zero) / total_val
    density = non_zero / total_val
    return f"Sparsity: {sparsity * 100:.4f}% | Density {density * 100:.4f}% | NNZ: {non_zero} | FMT: {fmt}"



"""Simple python ulitities for various projects."""
from importlib.metadata import PackageNotFoundError, version

from loguru import logger

try:
    __version__ = version("koyo")
except PackageNotFoundError:
    __version__ = "uninstalled"

__author__ = "Lukasz G. Migas"
__email__ = "lukas.migas@yahoo.com"


logger.disable("koyo")



"""PDF export mixin class."""

from __future__ import annotations

import typing as ty
from contextlib import contextmanager
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from loguru import logger
from PIL import Image

from koyo.typing import PathLike

if ty.TYPE_CHECKING:
    from matplotlib.backends.backend_pdf import PdfPages


class PDFMixin:
    """Mixin class to help export figures to PDF files."""

    _pdf = None
    as_pdf: bool = False

    # @property
    # def filename(self) -> str:
    #     """Filename that can be used for PDF generation."""
    #     raise NotImplementedError("Must implement method")

    @property
    def pdf_filename(self) -> Path:
        """Return the PDF filename."""
        raise NotImplementedError("Must implement method")
        # if self.filename is None:
        #     raise ValueError("PDF filename not set")
        # return Path(self.filename).with_suffix(".pdf")

    @property
    def pdf(self) -> PdfPages | None:
        """Return PDF figure."""
        if self._pdf is None and self.as_pdf:
            from matplotlib.backends.backend_pdf import PdfPages

            self._pdf = PdfPages(self.pdf_filename)
        return self._pdf

    @pdf.setter
    def pdf(self, value: PdfPages | None) -> None:
        """Set PDF figure."""
        if self._pdf is not None:
            self._pdf.close()
        self._pdf = value

    @staticmethod
    def _make_pdf(filename: PathLike) -> PdfPages:
        """Make PDF."""
        from matplotlib.backends.backend_pdf import PdfPages

        return PdfPages(filename)

    @contextmanager
    def _export_pdf_figures(self, filename: PathLike | None = None) -> ty.Generator[PdfPages | None, None, None]:
        """Export figures."""
        import matplotlib

        matplotlib.use("agg")

        pdf, reset = None, False
        if self.as_pdf:
            if filename:
                pdf = self._make_pdf(filename)
            else:
                pdf = self.pdf
                reset = True
        try:
            yield pdf
        except Exception as e:
            logger.error(f"Error exporting PDF: {e}")
        finally:
            if self.as_pdf:
                self._save_pdf(pdf, reset=reset)

    def _check_figure(self, filename: Path, override: bool = False) -> bool:
        """Check if figure exists.

        Will return True if figure should be created.
        """
        if self.as_pdf:
            return True
        return override or not filename.exists()

    def _add_title_to_pdf(self, title: str, pdf: PdfPages | None = None) -> None:
        pdf = pdf or self.pdf
        add_title_to_pdf(pdf, title)

    def _add_content_to_pdf(self, content: str, title: str = "", pdf: PdfPages | None = None) -> None:
        pdf = pdf or self.pdf
        add_content_to_pdf(pdf, content, title)

    def _save_pdf(self, pdf: PdfPages, reset: bool = False) -> None:
        """Save PPTX."""
        if pdf and hasattr(pdf, "close"):
            pdf.close()
        if pdf and hasattr(pdf, "filename"):
            Path(pdf.filename)
        logger.debug(f"Saved PDF to {self.pdf_filename}")
        if reset:
            self._pdf = None

    def _add_mpl_figure_to_pdf(
        self,
        filename: PathLike,
        fig: plt.Figure,
        face_color: str | np.ndarray | None = None,
        bbox_inches: str | None = "tight",
        dpi: int = 150,
        override: bool = False,
        if_empty: str = "warn",
        close: bool = False,
        pdf: PdfPages | None = None,
        **kwargs: ty.Any,
    ) -> None:
        """Export figure to file."""
        if fig is None:
            self._inform_on_empty(if_empty)
            return

        pdf = pdf or self.pdf
        add_mpl_figure_to_pdf(pdf, filename, fig, face_color, bbox_inches, dpi, override, close=close, **kwargs)

    def _add_pil_image_to_pdf(
        self,
        filename: PathLike,
        image: Image,
        override: bool = False,
        close: bool = False,
        pdf: PdfPages | None = None,
        **kwargs: ty.Any,
    ) -> None:
        """Export figure to file."""
        pdf = pdf or self.pdf
        add_pil_image_to_pdf(pdf, filename, image, override=override, close=close, **kwargs)

    @staticmethod
    def _inform_on_empty(if_empty: str = "warn") -> None:
        """Inform the user if the figure was empty."""
        if if_empty == "none":
            return
        elif if_empty == "warn":
            logger.warning("Figure was empty")
        elif if_empty == "raise":
            raise ValueError("Figure was empty")


def add_title_to_pdf(pdf: PdfPages | None, title: str) -> None:
    """Export title."""
    if title and pdf is not None:
        fig = plt.figure(figsize=(11.69, 8.27))
        fig.clf()
        fig.text(0.5, 0.5, title, transform=fig.transFigure, size=24, ha="center")
        pdf.savefig()
        plt.close(fig)


def add_content_to_pdf(pdf: PdfPages | None, content: str, title: str = "") -> None:
    """Export content."""
    if content and pdf is not None:
        fig = plt.figure(figsize=(11.69, 8.27))
        fig.clf()
        if title:
            fig.text(0.5, 0.9, title, transform=fig.transFigure, size=16, ha="center")
        fig.text(0.5, 0.5, content, transform=fig.transFigure, size=12, ha="center")


def add_mpl_figure_to_pdf(
    pdf: PdfPages | None,
    filename: PathLike,
    fig: plt.Figure,
    face_color: str | np.ndarray | None = None,
    bbox_inches: str | None = "tight",
    dpi: int = 150,
    override: bool = False,
    close: bool = False,
    format: str = "jpg",
    **kwargs: ty.Any,
) -> None:
    """Export figure to file."""
    face_color = face_color if face_color is not None else fig.get_facecolor()
    if pdf is not None:
        pdf.savefig(dpi=dpi, facecolor=face_color, bbox_inches=bbox_inches, **kwargs)
    else:
        if override or not Path(filename).exists():
            fig.savefig(filename, dpi=dpi, facecolor=face_color, bbox_inches=bbox_inches, format=format, **kwargs)
    if close:
        plt.close(fig)


def add_pil_image_to_pdf(
    pdf: PdfPages | None,
    filename: PathLike,
    image: Image,
    override: bool = False,
    dpi: int = 150,
    close: bool = False,
    **kwargs: ty.Any,
) -> None:
    """Export PIL image to PDF file (without closing it)."""
    if pdf is not None:
        fig = plt.figure(figsize=(image.width / 72, image.height / 72))
        ax = fig.add_axes([0, 0, 1, 1])
        ax.set_axis_off()
        ax.imshow(image, aspect="auto")
        pdf.savefig(fig, dpi=dpi, **kwargs)
        plt.close(fig)
    else:
        if override or not Path(filename).exists():
            image.save(filename, dpi=(dpi, dpi), **kwargs)
    if close:
        image.close()



"""Mosaic utilities."""

from __future__ import annotations

import io
import typing as ty
from functools import lru_cache
from math import ceil

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import to_rgba_array, ListedColormap
from tqdm import tqdm

from koyo.typing import PathLike
from koyo.utilities import rotate
from koyo.visuals import fix_style

if ty.TYPE_CHECKING:
    from PIL import Image


class RectanglePacker:
    """Rectangle packer."""

    def __init__(self, min_width: int = 0, x_pad: int = 0, y_pad: int = 0):
        self.free_rectangles = []  # List of free spaces (x, y, w, h)
        self.placed_rectangles = []  # List of placed rectangles (x, y, w, h)
        self.current_width = 0  # Current canvas width
        self.current_height = 0  # Current canvas height
        self.min_width = min_width
        self.x_pad = x_pad
        self.y_pad = y_pad

    def initialize_canvas(self, initial_width: int, initial_height: int):
        """Initialize the canvas with a starting size."""
        self.current_width = max(initial_width, self.min_width)
        self.current_height = initial_height
        self.free_rectangles = [(0, 0, self.current_width, self.current_height)]

    def place_rectangle(self, rect_width: int, rect_height: int) -> tuple[int, int]:
        """
        Place a rectangle of size (rect_width, rect_height) in the packing area.
        Returns the (x, y) position if placed, or raises an error if no space is available.
        """
        padded_width = rect_width + self.x_pad * 2
        padded_height = rect_height + self.y_pad * 2

        for i, (x, y, w, h) in enumerate(self.free_rectangles):
            if padded_width <= w and padded_height <= h:
                # Place the rectangle
                self.placed_rectangles.append((x + self.x_pad, y + self.y_pad, rect_width, rect_height))
                self.split_free_space(i, x, y, padded_width, padded_height)
                return x + self.x_pad, y + self.y_pad

        # If no space is found, expand the canvas
        self.expand_canvas(padded_width, padded_height)
        return self.place_rectangle(rect_width, rect_height)

    def expand_canvas(self, rect_width: int, rect_height: int):
        """Expand the canvas size to accommodate a rectangle that doesn't fit."""
        new_width = max(self.current_width, rect_width)
        new_height = self.current_height + rect_height
        self.free_rectangles.append((0, self.current_height, new_width, rect_height))
        self.current_width = new_width
        self.current_height = new_height

    def split_free_space(self, index: int, x: int, y: int, rect_width: int, rect_height: int):
        """Split the free space at index into smaller regions after placing a rectangle."""
        original_x, original_y, original_w, original_h = self.free_rectangles.pop(index)

        # Space to the right of the rectangle
        right_space = (x + rect_width, y, original_x + original_w - (x + rect_width), rect_height)
        if right_space[2] > 0 and right_space[3] > 0:
            self.free_rectangles.append(right_space)

        # Space below the rectangle
        below_space = (x, y + rect_height, original_w, original_y + original_h - (y + rect_height))
        if below_space[2] > 0 and below_space[3] > 0:
            self.free_rectangles.append(below_space)

        # Remaining space to the left and above (if applicable)
        left_space = (original_x, original_y, x - original_x, original_h)
        if left_space[2] > 0 and left_space[3] > 0:
            self.free_rectangles.append(left_space)

        above_space = (original_x, original_y, original_w, y - original_y)
        if above_space[2] > 0 and above_space[3] > 0:
            self.free_rectangles.append(above_space)

        # Sort free spaces by area (smaller spaces prioritized for tight packing)
        self.free_rectangles.sort(key=lambda r: r[2] * r[3])

    def get_packed_area(self) -> tuple[int, int]:
        """Returns the width and height of the smallest rectangle that fits all placed items."""
        max_width = max(x + w for x, y, w, h in self.placed_rectangles)
        max_height = max(y + h for x, y, w, h in self.placed_rectangles)
        return max(max_width, self.min_width), max_height


def get_positions(
    images: list[Image], min_width: int = 0, x_pad: int = 0, y_pad: int = 0
) -> tuple[tuple[int, int], list[tuple[int, int]], list[Image]]:
    """Get positions and canvas size."""
    packer = RectanglePacker(min_width=min_width, x_pad=x_pad, y_pad=y_pad)

    # Sort images by area (descending) for better packing
    images = sorted(images, key=lambda im: im.width * im.height, reverse=True)

    # Start with an initial canvas size
    initial_width = max(im.width for im in images) + x_pad * 2
    initial_height = sum(im.height for im in images) // len(images) + y_pad * 2
    packer.initialize_canvas(initial_width, initial_height)

    # Pack each image
    packed_positions = []
    for image in images:
        rect_width, rect_height = image.width, image.height
        position = packer.place_rectangle(rect_width, rect_height)
        packed_positions.append(position)

    # Get the final bounding box for the packed items
    packed_width, packed_height = packer.get_packed_area()
    return (packed_width, packed_height), packed_positions, images


def pack_images(images: list[Image], min_width: int = 0, x_pad: int = 0, y_pad: int = 0) -> Image:
    """Packs a list of images into a high-density rectangle packing with automatic canvas sizing."""
    packer = RectanglePacker(min_width=min_width, x_pad=x_pad, y_pad=y_pad)

    # Sort images by area (descending) for better packing
    images = sorted(images, key=lambda im: im.width * im.height, reverse=True)

    # Start with an initial canvas size
    initial_width = max(im.width for im in images) + x_pad * 2
    initial_height = sum(im.height for im in images) // len(images) + y_pad * 2
    packer.initialize_canvas(initial_width, initial_height)

    # Pack each image
    packed_positions = []
    for image in images:
        rect_width, rect_height = image.width, image.height
        position = packer.place_rectangle(rect_width, rect_height)
        packed_positions.append((image, position))

    # Get the final bounding box for the packed items
    packed_width, packed_height = packer.get_packed_area()

    # Create a canvas to draw the packed images
    canvas = Image.new("RGB", (packed_width, packed_height), (0, 0, 0))
    for image, (x, y) in packed_positions:
        canvas.paste(image, (x, y))
    return canvas


def add_label(
    ax: plt.Axes,
    label: str,
    x: float = 0.9,
    y: float = 0.98,
    color: str = "w",
    size: int = 14,
    weight: str = "normal",
    bbox: ty.Any = None,
    va: str = "top",
    ha: str = "left",
) -> plt.Text:
    """Add label to the image."""
    return ax.text(
        x,
        y,
        label,
        transform=ax.transAxes,
        fontsize=size,
        fontweight=weight,
        verticalalignment=va,
        horizontalalignment=ha,
        color=color,
        bbox=bbox,
    )


def fig_to_bytes(
    fig: plt.Figure | Image,
    bbox_inches: str | None = "tight",
    pad_inches: float = 0.1,
    dpi: int = 100,
    close: bool = False,
    transparent: bool = True,
) -> io.BytesIO:
    """Convert matplotlib figure to bytes."""
    buf = io.BytesIO()
    if isinstance(fig, plt.Figure):
        fig.savefig(
            buf,
            format="jpg",
            dpi=dpi,
            facecolor=fig.get_facecolor(),
            edgecolor=fig.get_edgecolor(),
            bbox_inches=bbox_inches,
            pad_inches=pad_inches,
            transparent=transparent,
        )
        if close:
            plt.close(fig)
    else:
        fig.save(buf, format="PNG")
        if close:
            fig.close()
    buf.seek(0)
    return buf


def make_fig_title(title: str, width: int, n_cols: int, text_color: str = "auto", font_size: int = 36) -> io.BytesIO:
    """Make title for a single image."""
    n_rows = max(1, title.count("\n"))
    fig = plt.figure(figsize=(width * n_cols / 72, n_rows))
    text_color = text_color if text_color != "auto" else plt.rcParams["text.color"]

    fig.text(
        0.5,
        0.5,
        title,
        transform=fig.transFigure,
        size=font_size,
        ha="center",
        va="center",
        color=text_color,
    )
    return fig_to_bytes(fig, bbox_inches=None, pad_inches=0, dpi=72, close=True)


def merge_mosaic(
    items: dict[str, io.BytesIO],
    title_buf: io.BytesIO | None = None,
    title: str = "",
    n_cols: int | None = None,
    silent: bool = True,
    color: tuple[int, ...] = (0, 0, 0, 0),  # black
    allow_placeholder: bool = False,
    placeholder_color: tuple[int, ...] = (0, 0, 0, 255),  # black
    x_pad: int = 0,
    y_pad: int = 0,
    add_title_to_images: bool = False,
    text_color: str | tuple[int, int, int, int] = (255, 255, 255, 255),
    font_size: int = 36,
) -> Image:
    """Merge images."""
    nr, nc, w, h = _get_mosaic_dims_for_list(items, n_cols=n_cols, x_pad=x_pad, y_pad=y_pad)
    if title:
        title_buf = make_fig_title(title, w, nc, font_size=font_size)
    return _merge_mosaic(
        nr,
        nc,
        w,
        h,
        items,
        title_buf,
        silent,
        color,
        allow_placeholder,
        placeholder_color,
        x_pad=x_pad,
        y_pad=y_pad,
        add_title_to_images=add_title_to_images,
        text_color=text_color,
    )


def merge_mosaic_packed(
    items: dict[str, io.BytesIO],
    title_buf: io.BytesIO | None = None,
    title: str = "",
    min_width: int = 0,
    silent: bool = True,
    color: tuple[int, ...] = (0, 0, 0, 0),  # black
    allow_placeholder: bool = False,
    placeholder_color: tuple[int, ...] = (0, 0, 0, 255),  # black
    x_pad: int = 0,
    y_pad: int = 0,
):
    """Pack images tightly into a grid."""
    from PIL import Image

    images = [Image.open(buf) for buf in items.values()]
    (w, h), positions, images = get_positions(images, min_width=min_width, x_pad=x_pad, y_pad=y_pad)
    if title:
        title_buf = make_fig_title(title, w, 1)

    if title_buf is not None:
        title = Image.open(title_buf)
        dst = Image.new("RGB", (w, h + title.height), color=color)
        dst.paste(title, (0, 0))
        title.close()
        del title_buf
        y_offset = title.height
    else:
        dst = Image.new("RGB", (w, h), color=color)
        y_offset = 0
    with tqdm(desc="Merging images...", total=len(images), disable=silent) as pbar:
        for im, (x, y) in zip(images, positions):
            dst.paste(im, (x, y + y_offset))
            pbar.update(1)
    return dst


def merge_mosaic_from_dir(
    image_dir: PathLike,
    title: str = "",
    n_cols: int | None = None,
    x_pad: int = 0,
    y_pad: int = 0,
) -> Image:
    """Merge images from a directory."""
    items = {}
    for filename in image_dir.glob("*"):
        if filename.is_file():
            with open(filename, "rb") as f:
                items[filename.stem] = io.BytesIO(f.read())
    return merge_mosaic(items, title=title, n_cols=n_cols, x_pad=x_pad, y_pad=y_pad)


def merge_mosaic_with_columns(
    image_dir: PathLike | list[io.BytesIO],
    n_cols: int,
    x_pad: int = 0,
    y_pad: int = 0,
    placeholder_color: tuple[int, ...] = (0, 0, 0, 255),  # black
) -> Image:
    """Merge images from a directory into a grid with a fixed number of columns."""
    # Load images into a list
    from PIL import Image

    items = []
    if isinstance(image_dir, list):
        for buf in image_dir:
            with Image.open(buf) as im:
                items.append((im.copy(), im.width, im.height))
    else:
        for filename in image_dir.glob("*"):
            if filename.is_file():
                with open(filename, "rb") as f:
                    with Image.open(io.BytesIO(f.read())) as im:
                        items.append((im.copy(), im.width, im.height))

    if not items:
        raise ValueError("No images found in the specified directory.")

    # Calculate the number of rows
    num_images = len(items)
    n_rows = ceil(num_images / n_cols)

    # Determine the width of each column and height of each row
    column_widths = [0] * n_cols
    row_heights = [0] * n_rows

    # Assign images to the grid
    grid = [[None for _ in range(n_cols)] for _ in range(n_rows)]
    for idx, (im, width, height) in enumerate(items):
        row = idx // n_cols
        col = idx % n_cols
        grid[row][col] = im
        column_widths[col] = max(column_widths[col], width)
        row_heights[row] = max(row_heights[row], height)

    # Calculate the total mosaic dimensions
    mosaic_width = sum(column_widths) + (n_cols - 1) * x_pad
    mosaic_height = sum(row_heights) + (n_rows - 1) * y_pad

    # Create the output mosaic
    mosaic = Image.new("RGB", (mosaic_width, mosaic_height), color=placeholder_color)

    # Place images in the mosaic
    y_offset = 0
    for row_idx, row in enumerate(grid):
        x_offset = 0
        for col_idx, im in enumerate(row):
            if im:
                mosaic.paste(im, (x_offset, y_offset))
            x_offset += column_widths[col_idx] + x_pad
        y_offset += row_heights[row_idx] + y_pad

    return mosaic


def _merge_mosaic(
    n_rows: int,
    n_cols: int,
    width: int,
    height: int,
    items: dict[str, io.BytesIO],
    title_buf: io.BytesIO | None = None,
    silent: bool = True,
    color: tuple[int, ...] = (0, 0, 0, 0),
    allow_placeholder: bool = False,
    placeholder_color: tuple[int, ...] = (0, 0, 0, 255),
    x_pad: int = 0,
    y_pad: int = 0,
    add_title_to_images: bool = False,
    text_color: str | tuple[int, int, int, int] = (255, 255, 255, 255),
    align_to: ty.Literal["top", "center", "bottom"] = "center",
) -> Image:
    from PIL import Image

    if isinstance(color, tuple):
        assert len(color) in [3, 4], f"Color must be a tuple of 3 or 4 integers, not {color}."
    color = (*tuple(color), 255) if len(color) == 3 else color
    if isinstance(placeholder_color, tuple):
        assert len(placeholder_color) in [3, 4], (
            f"Placeholder color must be a tuple of 3 or 4 integers, not {placeholder_color}."
        )
    placeholder_color = (*tuple(placeholder_color), 255) if len(placeholder_color) == 3 else placeholder_color

    names = list(items.keys())
    filelist = list(items.values())
    if title_buf is not None:
        title = Image.open(title_buf)
        dst = Image.new("RGB", (width * n_cols, height * n_rows + title.height), color=color)
        dst.paste(title, (0, 0))
        title.close()
        del title_buf
        y_offset = title.height
    else:
        dst = Image.new("RGB", (width * n_cols, height * n_rows), color=color)
        y_offset = 0
    k = 0  # image counter
    with tqdm(desc="Merging images...", total=len(filelist), disable=silent) as pbar:
        for i in range(n_rows):  # iterate over rows
            for j in range(n_cols):  # iterate over columns
                try:
                    filename = filelist[k]
                    im = None
                    if filename:
                        im = Image.open(filename)
                    elif filename is None and allow_placeholder:
                        im = Image.new("RGB", (width, height), color=placeholder_color)

                    try:
                        if align_to == "top":
                            y = height * i + y_offset + y_pad
                        elif align_to == "center":
                            y = height * i + y_offset + (height - im.height) // 2
                        elif align_to == "bottom":
                            y = height * i + y_offset + height - im.height - y_pad
                    except AttributeError:
                        y = height * i + y_offset + y_pad

                    if im:
                        # load image
                        x = width * j + x_pad
                        dst.paste(im, (x, y))
                        if add_title_to_images and filename:
                            _add_text_to_pil_image(dst, str(names[k]), (x, y), text_color)
                        im.close()
                    k += 1  # increment image counter
                    pbar.update(1)
                except IndexError:
                    break
    return dst


def _get_mosaic_dims_for_list(
    items: dict[str, io.BytesIO],
    n_cols: int | None = 0,
    check_size_of_all: bool = True,
    x_pad: int = 0,
    y_pad: int = 0,
) -> tuple[int, int, int, int]:
    from PIL import Image

    if n_cols is None:
        n_cols = 0

    x_pad = x_pad * 2
    y_pad = y_pad * 2

    widths, heights = [], []
    if check_size_of_all:
        for buf in items.values():
            if buf is None:
                continue
            with Image.open(buf) as im:
                widths.append(im.width + x_pad)
                heights.append(im.height + y_pad)
        widths = np.unique(widths)
        heights = np.unique(heights)
        width, height = np.max(widths), np.max(heights)
    else:
        buf = next(iter(items.values()))
        with Image.open(buf) as im:
            width, height = im.width, im.height
    return _get_mosaic_dims(len(items), width, height, n_cols=n_cols)


def _get_mosaic_dims(n: int, width: int, height: int, n_cols: int = 0) -> tuple[int, int, int, int]:
    ratio = 100.0
    _width, _height = width, height
    if n_cols == 0:
        n_rows, n_cols = 1, 1
        desired_ratio = 16 / 9
        while ratio > desired_ratio:
            n_cols = ceil(n / n_rows)
            height = _height * n_rows
            ratio = (_width * n_cols) / height
            n_rows += 1
    else:
        if n_cols > n:
            n_cols = n
        n_rows = ceil(n / n_cols)
    while n_rows > ceil(n / n_cols):
        n_rows -= 1
    return n_rows, n_cols, _width, _height


def plot_mosaic(
    data: dict[str, np.ndarray],
    title: str = "",
    colormap: str | dict[str, str] = "viridis",
    colorbar: bool = True,
    dpi: int = 100,
    min_val: float | None = None,
    max_val: float | None = None,
    figsize: tuple[float, float] = (6, 6),
    n_cols: int | None = None,
    style: str = "dark_background",
    color: tuple[int, int, int, int] | None = None,
    placeholder_color: tuple[int, int, int, int] | None = None,
    highlight: str | None = None,
    auto_rotate: bool = False,
    **kwargs: ty.Any,
) -> Image:
    """Plot mosaic."""
    from koyo.visuals import _plot_or_update_image

    img, cbar = None, None
    figures = {}
    with plt.style.context(fix_style(style)):
        if color is None:
            color = _get_color_rgba255(plt.rcParams["axes.facecolor"])
        if placeholder_color is None:
            placeholder_color = _get_color_rgba255(plt.rcParams["axes.facecolor"])

        fig, ax = plt.subplots(figsize=figsize)
        ax.axis("off")

        for key in data:
            img, cbar = _plot_or_update_image(
                ax,
                rotate(data[key], auto_rotate),
                min_val=min_val,
                max_val=max_val,
                img=img,
                cbar=cbar,
                colorbar=colorbar,
                colormap=colormap if isinstance(colormap, str) else colormap[key],
                title=key,
            )
            if highlight and key == highlight:
                # change ax title color
                ax.title.set_color("red")
            else:
                ax.title.set_color(plt.rcParams["text.color"])
            figures[key] = fig_to_bytes(fig, close=False, dpi=dpi)
        plt.close(fig)
        image = merge_mosaic(
            figures,
            title=title,
            n_cols=n_cols,
            placeholder_color=placeholder_color,
            color=color,
        )
    return image


def plot_mosaic_no_colorbar(
    data: dict[str, np.ndarray],
    colormap: str | dict[str, str] = "viridis",
    min_val: float | None = None,
    max_val: float | None = None,
    title: str = "",
    auto_rotate: bool = False,
    n_cols: int | None = None,
    style: str = "dark_background",
    color: tuple[int, int, int, int] | None = None,
    placeholder_color: tuple[int, int, int, int] | None = None,
    **kwargs: ty.Any,
) -> Image:
    """Plot mosaic without colorbar."""
    images = {}
    with plt.style.context(style):
        if color is None:
            color = _get_color_rgba255(plt.rcParams["axes.facecolor"])
        if placeholder_color is None:
            placeholder_color = _get_color_rgba255(plt.rcParams["axes.facecolor"])
        for key in data:
            images[key] = fig_to_bytes(
                convert_array_to_image(
                    rotate(data[key], auto_rotate),
                    title="",
                    min_val=min_val,
                    max_val=max_val,
                    colormap=colormap if isinstance(colormap, str) else colormap[key],
                )
            )
        image = merge_mosaic(
            images,
            title=title,
            x_pad=5,
            y_pad=5,
            add_title_to_images=True,
            n_cols=n_cols,
            placeholder_color=placeholder_color,
            color=color,
        )
    return image


def plot_mosaic_individual(
    data: dict[str, np.ndarray],
    title: str = "",
    colormap: str | dict[str, str] = "viridis",
    colorbar: bool = True,
    dpi: int = 100,
    min_val: float | None = None,
    max_val: float | None = None,
    figsize: tuple[float, float] = (6, 6),
    style: str = "dark_background",
    n_cols: int | None = None,
    border_color: dict[str, str] | None = None,
    title_color: dict[str, str] | None = None,
    color: tuple[int, int, int, int] | None = None,
    placeholder_color: tuple[int, int, int, int] | None = None,
    auto_rotate: bool = False,
    **kwargs: ty.Any,
) -> Image:
    """Plot mosaic."""
    from koyo.visuals import _plot_image

    border_color = {} if border_color is None else border_color
    title_color = {} if title_color is None else title_color

    figures = {}
    with plt.style.context(fix_style(style)):
        if color is None:
            color = _get_color_rgba255(plt.rcParams["axes.facecolor"])
        if placeholder_color is None:
            placeholder_color = _get_color_rgba255(plt.rcParams["axes.facecolor"])

        for key in data:
            fig, ax = _plot_image(
                rotate(data[key], auto_rotate),
                min_val=min_val,
                max_val=max_val,
                colorbar=colorbar,
                colormap=colormap if isinstance(colormap, str) else colormap[key],
                title=key,
                figsize=figsize,
                border_color=border_color.get(key, None),
                title_color=title_color.get(key, None),
            )
            ax.axis("off")
            figures[key] = fig_to_bytes(fig, close=True, dpi=dpi)
        image = merge_mosaic(
            figures,
            title=title,
            n_cols=n_cols,
            placeholder_color=placeholder_color,
            color=color,
        )
    return image


def plot_mosaic_with_mask_individual(
    data: dict[str, np.ndarray],
    mask: dict[str, np.ndarray] | None = None,
    mask_color: str = "red",
    mask_opacity: float = 0.3,
    title: str = "",
    colormap: str | dict[str, str] = "viridis",
    colorbar: bool = True,
    dpi: int = 100,
    min_val: float | None = None,
    max_val: float | None = None,
    figsize: tuple[float, float] = (6, 6),
    style: str = "dark_background",
    n_cols: int | None = None,
    border_color: dict[str, str] | None = None,
    title_color: dict[str, str] | None = None,
    color: tuple[int, int, int, int] | None = None,
    placeholder_color: tuple[int, int, int, int] | None = None,
    auto_rotate: bool = False,
    **kwargs: ty.Any,
) -> Image:
    """Plot mosaic."""
    from koyo.visuals import _plot_image

    border_color = {} if border_color is None else border_color
    title_color = {} if title_color is None else title_color

    if not mask:
        mask = {}

    # create mask colormap
    mask_colormap = ListedColormap([mask_color])
    mask_colormap.set_bad(alpha=0.)
    mask_colormap.set_over(alpha=0.)
    mask_colormap.set_under(alpha=0.)

    figures = {}
    with plt.style.context(fix_style(style)):
        if color is None:
            color = _get_color_rgba255(plt.rcParams["axes.facecolor"])
        if placeholder_color is None:
            placeholder_color = _get_color_rgba255(plt.rcParams["axes.facecolor"])

        for key in data:
            fig, ax = _plot_image(
                rotate(data[key], auto_rotate),
                min_val=min_val,
                max_val=max_val,
                colorbar=colorbar,
                colormap=colormap if isinstance(colormap, str) else colormap[key],
                title=key,
                figsize=figsize,
                border_color=border_color.get(key, None),
                title_color=title_color.get(key, None),
            )
            if key in mask:
                masked = np.ma.masked_where(mask[key] == 0, mask[key])
                ax.imshow(masked, cmap=mask_colormap, alpha=mask_opacity, interpolation='none')
            ax.axis("off")
            figures[key] = fig_to_bytes(fig, close=True, dpi=dpi)
        image = merge_mosaic(
            figures,
            title=title,
            n_cols=n_cols,
            placeholder_color=placeholder_color,
            color=color,
        )
    return image


def plot_mosaic_line_individual(
    data: dict[str, tuple[np.ndarray, np.ndarray]],
    marker: float | dict[str, float] | None = None,
    x_label: str = "",
    y_label: str = "",
    title: str = "",
    dpi: int = 100,
    figsize: tuple[float, float] = (6, 6),
    style: str = "seaborn-ticks",
    n_cols: int | None = None,
    border_color: dict[str, str] | None = None,
    title_color: dict[str, str] | None = None,
    color: tuple[int, int, int, int] | None = None,
    placeholder_color: tuple[int, int, int, int] | None = None,
) -> Image:
    """Plot mosaic."""
    from koyo.visuals import _plot_line

    border_color = {} if border_color is None else border_color
    title_color = {} if title_color is None else title_color

    figures = {}
    with plt.style.context(fix_style(style)):
        if color is None:
            color = _get_color_rgba255(plt.rcParams["axes.facecolor"])
        if placeholder_color is None:
            placeholder_color = _get_color_rgba255(plt.rcParams["axes.facecolor"])
        for key in data:
            fig, ax = _plot_line(
                data[key][0],
                data[key][1],
                marker=marker[key] if isinstance(marker, dict) else marker,
                title=key,
                x_label=x_label,
                y_label=y_label,
                figsize=figsize,
                border_color=border_color.get(key, None),
                title_color=title_color.get(key, None),
            )
            figures[key] = fig_to_bytes(fig, close=True, dpi=dpi)
        image = merge_mosaic(
            figures,
            title=title,
            n_cols=n_cols,
            color=color,
            placeholder_color=placeholder_color,
        )
    return image


def _get_color_rgba255(color) -> tuple[int, int, int, int]:
    color = to_rgba_array(color) * 255
    color = color.astype(int)
    if color.ndim == 2:
        return tuple(color[0])
    return tuple(color)


@lru_cache(maxsize=10)
def find_font(name: str = "arial.ttf") -> str | None:
    """Find font."""
    # get list of fonts
    from matplotlib import font_manager

    fonts = font_manager.findSystemFonts(fontpaths=None, fontext="ttf")
    for font in fonts:
        if name in font.lower():
            return font
    return None


def convert_array_to_image(
    array: np.ndarray,
    colormap: str = "viridis",
    title: str = "",
    min_val: float | None = None,
    max_val: float | None = None,
) -> Image:
    """Convert a 2D array to a PNG image using a specified colormap."""
    from matplotlib.colors import Normalize, to_hex
    from PIL import Image

    # Define the colormap
    cmap = plt.get_cmap(colormap)

    # Normalize the array excluding NaNs or zeros
    valid_mask = ~np.isnan(array) & (array != 0)
    min_val = min_val if min_val is not None else np.min(array[valid_mask])
    max_val = max_val if max_val is not None else np.max(array[valid_mask])
    norm = Normalize(vmin=min_val, vmax=max_val)

    # Create the normalized array
    normalized_array = np.zeros_like(array, dtype=float)  # Default to 0 for NaNs and zeros
    normalized_array[valid_mask] = norm(array[valid_mask])

    # Apply the colormap
    colored_image = cmap(normalized_array)

    # Set NaNs and zeros to black (0, 0, 0, 1 in RGBA)
    colored_image[~valid_mask] = (0, 0, 0, 1)

    # Convert to 8-bit unsigned integers
    uint8_image = (colored_image[:, :, :3] * 255).astype(np.uint8)  # Drop alpha channel for saving

    # Save the image
    image = Image.fromarray(uint8_image)
    if title:
        if colormap == "viridis":
            color = "white"
        else:
            color = to_hex(plt.rcParams["text.color"])
        _add_text_to_pil_image(image, title, (3, 3), color)
    return image


def _add_text_to_pil_image(
    image: Image, text: str, position: tuple[int, int], color: str | tuple[int, int, int, int]
) -> None:
    from PIL import ImageDraw, ImageFont

    font_path = find_font()
    if font_path:
        font = ImageFont.truetype(font_path, 14)
    else:
        font = ImageFont.load_default()
    draw = ImageDraw.Draw(image)
    draw.text(position, text, fill=color, font=font)  # (255, 255, 255, 255))


def export_array_as_bytes(array: np.ndarray, colormap: str = "viridis") -> io.BytesIO:
    """Export a 2D array as a PNG image in bytes format."""
    image = convert_array_to_image(array, colormap)
    output = io.BytesIO()
    image.save(output, format="PNG")
    return output



import numpy as np
import pytest
from koyo.image import clip_hotspots, pearson_similarity


@pytest.mark.parametrize("quantile", (0.99, 0.75))
def test_clip_hotspots(quantile):
    img = np.random.randint(0, 255, (50, 50))

    _img = clip_hotspots(img, quantile)
    np.testing.assert_array_equal(img.shape, _img.shape)


@pytest.mark.parametrize("size", ((3, 3), (5, 5)))
def test_pearson_similarity(size):
    img = np.random.randint(0, 255, (50, 50))

    score = pearson_similarity(img, img, size)
    assert score == 1.0

    with pytest.raises(ValueError):
        pearson_similarity(img, img, (3, 3, 3))



"""Generic utilities."""

from __future__ import annotations

import re
import typing as ty
from collections.abc import Iterable
from math import ceil, floor
from pathlib import Path
from random import shuffle

import numba as nb
import numpy as np
from natsort import natsorted
from loguru import logger

from koyo.typing import PathLike, SimpleArrayLike
from koyo.system import running_as_pyinstaller_app

if ty.TYPE_CHECKING:
    from koyo.fig_mixin import PptxPdfWrapper


def human_readable_byte_size(n_bytes: int) -> str:
    """

    Parameters
    ----------
    n_bytes : int
        Number of bytes.

    Returns
    -------
    Human readable string : str

    """
    suffixes = ["B", "KB", "MB", "GB", "TB", "PB"]
    i = 0
    while n_bytes >= 1024 and i < len(suffixes) - 1:
        n_bytes /= 1024.0
        i += 1
    f = (f"{n_bytes:.2f}").rstrip("0").rstrip(".")
    return f"{f} {suffixes[i]}"


def split_tooltip(text: str, max_width: int = 80) -> str:
    """Split tooltip text into multiple lines based on maximum width.

    Args:
        text: Text to be split
        max_width: Maximum number of characters per line (default: 80)

    Returns:
        Formatted string with line breaks
    """
    if not text:
        return ""

    words = text.split()
    lines = []
    current_line = []
    current_width = 0
    for word in words:
        word_len = len(word)
        if current_width + word_len + 1 <= max_width:  # +1 for space
            current_line.append(word)
            current_width += word_len + 1
        else:
            if current_line:
                lines.append(" ".join(current_line))
            current_line = [word]
            current_width = word_len + 1

    if current_line:
        lines.append(" ".join(current_line))
    return "\n".join(lines)


def get_unique_without_sort(seq: list) -> list:
    """Get unique values without sorting."""
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


def ensure_output_dir_exists(
    output_dir: PathLike | None, pptx_or_pdf: PptxPdfWrapper | None = None
) -> tuple[bool, Path]:
    """Ensure that the output directory exists.

    Returns
    -------
    bool
        whether the directory was created
    Path
        output directory
    """
    # check if the output directory is actually needed - not if exporting as PDF or PPTX
    if pptx_or_pdf and pptx_or_pdf.as_pptx_or_pdf:
        return False, Path("")
    if output_dir:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        return True, output_dir
    return False, Path("")


def clean_path(path: str) -> Path:
    """Clean path that might have been stored in a wrong way."""
    path = str(path)
    # replace Windows backslashes with forward slashes
    path = path.replace("\\", "/")
    # replace MacOS network path with local path and resolve the network drive if it exists
    path = path.replace("/Volumes/", "/")
    return Path(path)


def is_installed(module: str) -> bool:
    """Try to import module."""
    import importlib.util

    try:
        loader = importlib.util.find_spec(module)
    except ModuleNotFoundError:
        return False
    return loader is not None


def get_version(module: str) -> str:
    """Get current version of package."""
    import importlib.metadata

    try:
        installed_version = importlib.metadata.version(module)
    except importlib.metadata.PackageNotFoundError:
        return "N/A"
    return installed_version


def is_above_version(module: str, version: str) -> bool:
    """Check whether the module is above a certain version."""
    import importlib.metadata

    from packaging.version import Version

    try:
        installed_version = importlib.metadata.version(module)
    except importlib.metadata.PackageNotFoundError:
        logger.warning(f"Module {module} not found.")
        return False
    installed_version = Version(installed_version)
    version = Version(version)
    if installed_version is not None:
        return installed_version >= version
    logger.warning(f"Module {module} not found.")
    return False


def get_format(fmt: str) -> str:
    """Parse format."""
    return fmt if fmt.startswith(".") else f".{fmt}"


def ensure_list(array: np.ndarray | list[np.ndarray]) -> list[np.ndarray]:
    """Ensure that the input is a list."""
    if isinstance(array, np.ndarray):
        return [array]
    return array


def find_nearest_divisor(
    value: int | float,
    divisor: int | float = 1,
    increment: int | float = 1,
    max_iters: int = 1000,
) -> int | float:
    """Find nearest divisor.

    Parameters
    ----------
    value : Union[int, float]
        value to be divided by the divisor
    divisor : Union[int, float]
        value by which the `value` is divided by
    increment : Union[int, float]
        increment by which to change the `divisor`
    max_iters : int
        maximum number of iterations before the algorithm should give up

    Returns
    -------
    divisor : Union[int, float]
        divisor value or -1 if the algorithm did not find appropriate value

    """
    if divisor > value:
        raise ValueError("Initial guess cannot be larger than value")

    n_iter = 0
    while value % divisor != 0 and n_iter < max_iters:
        divisor += increment
        n_iter += 1

    if value % divisor == 0:
        return divisor
    return -1


def slugify_name(value: str) -> str:
    """Slugify filename."""
    return value.replace(" ", "_").replace("/", "_").replace("\\", "_").replace(":", "-")


def order_parameters(**kwargs: ty.Any) -> dict:
    """Order parameters."""
    kwargs_ = {}
    for key in natsorted(kwargs):
        kwargs_[key] = kwargs[key]
    return kwargs_


def exclude_parameters(exclude: ty.Iterable[str], **kwargs: ty.Any) -> dict:
    """Exclude parameters."""
    kwargs_ = {}
    for key in natsorted(kwargs):
        if key not in exclude:
            kwargs_[key] = kwargs[key]
    return kwargs_


def get_module_path(module: str, filename: str) -> str:
    """Get module path."""
    import importlib.resources

    if not filename.endswith(".py"):
        filename += ".py"

    path = str(importlib.resources.files(module).joinpath(filename))
    return path


def reraise_exception_if_debug(exc, message: str = "Exception occurred", env_key: str = "DEV_MODE") -> None:
    """Reraise exception if debug mode is enabled and jump into the debugger."""
    import os

    from loguru import logger

    if os.environ.get(env_key, "0") == "1":
        raise exc
    logger.exception(message)


def pluralize(word: str, n: int, with_e: bool = False) -> str:
    """Give plural form to a word."""
    if word in ["is", "are"]:
        return "is" if n == 1 else "are"
    if word in ["was", "were"]:
        return "was" if n == 1 else "were"
    elif word in ["has", "have"]:
        return "has" if n == 1 else "have"
    extra = "s" if not with_e else "es"
    return word if n == 1 else word + extra


def is_valid_python_name(name: str) -> bool:
    """Check whether name is valid."""
    from keyword import iskeyword

    return name.isidentifier() and not iskeyword(name)


def is_between(value: float, lower: float, upper: float, inclusive: bool = True) -> bool:
    """Check if value is between lower and upper."""
    if inclusive:
        return lower <= value <= upper
    return lower < value < upper


def flatten_nested_list(list_of_lists: list[list]) -> list:
    """Flatten nested list of lists."""
    return [item for sublist in list_of_lists for item in sublist]


def get_list_difference(li1: list, li2: list) -> list:
    """Get difference between two lists."""
    # get difference between two lists while keeping the order
    li_dif = [i for i in li1 + li2 if i not in li1 or i not in li2]
    return li_dif


def find_nearest_index(data: SimpleArrayLike, value: int | (float | (np.ndarray | Iterable))):
    """Find nearest index of asked value.

    Parameters
    ----------
    data : np.array
        input array (e.g. m/z values)
    value : Union[int, float, np.ndarray]
        asked value

    Returns
    -------
    index :
        index value
    """
    data = np.asarray(data)
    if isinstance(value, Iterable):
        return np.asarray([np.argmin(np.abs(data - _value)) for _value in value], dtype=np.int64)
    return np.argmin(np.abs(data - value))


@nb.njit()
def find_nearest_index_array(data: SimpleArrayLike, value: np.ndarray | ty.Iterable) -> np.ndarray:
    """Find nearest index of asked value.

    Parameters
    ----------
    data : ArrayLike
        input array (e.g. m/z values)
    value : Union[int, float, np.ndarray]
        asked value

    Returns
    -------
    index :
        index value
    """
    data = np.asarray(data)
    return np.asarray([np.argmin(np.abs(data - _value)) for _value in value])


@nb.njit()
def find_nearest_index_single(data: SimpleArrayLike, value: int | float):
    """Find nearest index of asked value.

    Parameters
    ----------
    data : ArrayLike
        input array (e.g. m/z values)
    value : Union[int, float, np.ndarray]
        asked value

    Returns
    -------
    index :
        index value
    """
    return np.argmin(np.abs(data - value))


def find_nearest_value_single(data: SimpleArrayLike, value: int | float) -> int | float:
    """Find nearest value."""
    data = np.asarray(data)
    idx = find_nearest_index_single(data, value)
    return data[idx]


def find_nearest_value_in_dict(data: dict[float, ty.Any], value: int | float) -> ty.Any:
    """Find nearest value in dictionary."""
    key = find_nearest_value(list(data.keys()), value)
    return data[key]


def find_nearest_index_batch(array: SimpleArrayLike, values: SimpleArrayLike, sort: bool = False) -> np.ndarray:
    """Find nearest index."""
    # make sure array is a numpy array
    array = np.asarray(array)
    values = np.asarray(values)
    if not array.size or not values.size:
        return np.array([], dtype=np.int64)

    # get insert positions
    sorter = np.argsort(array) if sort else None
    indices = np.searchsorted(array, values, side="left", sorter=sorter)

    # find indexes where previous index is closer
    prev_idx_is_less = (indices == len(array)) | (
        np.fabs(values - array[np.maximum(indices - 1, 0)])
        < np.fabs(values - array[np.minimum(indices, len(array) - 1)])
    )
    indices[prev_idx_is_less] -= 1
    if sorter is not None:
        indices = sorter[indices]
    return indices


def find_nearest_value(data: ty.Iterable, value: int | (float | (np.ndarray | Iterable))):
    """Find nearest value."""
    data = np.asarray(data)
    idx = find_nearest_index(data, value)
    return data[idx]


def get_pixels_within_radius(grid: np.ndarray, x: int, y: int, radius: int = 0) -> np.ndarray:
    """Get pixels within radius."""
    # Get the bounds for the sub-grid (handling edge cases)
    y_min = max(0, y - radius)
    y_max = min(grid.shape[1], y + radius + 1)
    x_min = max(0, x - radius)
    x_max = min(grid.shape[0], x + radius + 1)

    # Extract the sub-grid around the point (y, x)
    sub_grid = grid[y_min:y_max, x_min:x_max].flatten()
    # drop NaNs, Infs, etc
    return sub_grid[np.isfinite(sub_grid)]


def get_kws(func: ty.Callable, **kwargs: ty.Any) -> dict:
    """Get kwargs."""
    import inspect

    args = inspect.getfullargspec(func).args

    kws = {}
    for kw in args:
        if kw in kwargs:
            kws[kw] = kwargs[kw]
    return kws


def format_count(value: float) -> str:
    """Format count."""
    if value < 1e3:
        return f"{value:.0f}"
    elif value < 1e6:
        return f"{value / 1e3:.1f}K"
    elif value < 1e9:
        return f"{value / 1e6:.1f}M"
    return f"{value / 1e9:.1f}B"


def format_size(size: int) -> str:
    """Convert bytes to nicer format."""
    if size < 2**10:
        return f"{size}"
    elif size < 2**20:
        return "%.1fK" % (size / float(2**10))
    elif size < 2**30:
        return "%.1fM" % (size / float(2**20))
    elif size < 2**40:
        return "%.1fG" % (size / float(2**30))
    elif size < 2**50:
        return "%.1fT" % (size / float(2**40))
    return "%.1fP" % (size / float(2**50))


def is_number(value: ty.Any) -> bool:
    """Quick and easy way to check if input is a number.

    Parameters
    ----------
    value : any
        input value

    Returns
    -------
    bool
        returns True/False, depending if a value is a number
    """
    return isinstance(value, (int, float, complex))


def check_value_order(value_min: float, value_max: float) -> tuple[float, float]:
    """Check whether the value order is correct (min -> max).

    Parameters
    ----------
    value_min : int or float or complex
        presumed minimal value
    value_max : int or float or complex
        presumed maximal value

    Returns
    -------
    value_min : int or float or complex
        true minimal value
    value_max : int or float or complex
        true maximal value
    """
    if not is_number(value_min) or not is_number(value_max):
        return value_min, value_max

    if value_max < value_min:
        value_max, value_min = value_min, value_max
    return value_min, value_max


def get_value(new_value: ty.Any, current_value: ty.Any) -> ty.Any:
    """Get value."""
    if new_value is None:
        return current_value
    return new_value


def rescale(
    values: np.ndarray | list,
    new_min: float,
    new_max: float,
    dtype=None,
    min_val: float | None = None,
    max_val: float | None = None,
) -> np.ndarray:
    """Rescale values from one range to another.

    Parameters
    ----------
    values : Union[np.ndarray, List]
        input range
    new_min : float
        new minimum value
    new_max : float
        new maximum value
    dtype :
        data type
    min_val: float
        minimum value of the original range
    max_val: float
        maximum value of the original range

    Returns
    -------
    new_values : np.ndarray
        rescaled range
    """
    values = np.asarray(values)
    if dtype is None:
        dtype = values.dtype
    old_min, old_max = get_min_max(values)
    if min_val is not None:
        old_min = min([min_val, old_min])
    if max_val is not None:
        old_max = max([max_val, old_max])
    # check if dtype is integer and new dtype is float, in this case, you can cast it
    if np.issubdtype(dtype, np.integer) and np.issubdtype(values.dtype, np.floating):
        values = values.astype(dtype)
    # actually rescale
    new_values = ((values - old_min) / (old_max - old_min)) * (new_max - new_min) + new_min
    return new_values.astype(dtype)


def rescale_value(
    value: float,
    old_min: float,
    old_max: float,
    new_min: float = 0.0,
    new_max: float = 1.0,
    clip: bool = True,
) -> float:
    """Rescale value to new range.

    Parameters
    ----------
    value : float
        value that needs to be rescaled
    old_min : float
        minimum value of the original range
    old_max : float
        maximum value of the original range
    new_min : float
        minimum value of the new range
    new_max : float
        maximum value of the new range
    clip : bool
        if True, the new value will be clipped to fit inside of the new maximum range
    """
    if clip:
        if value < old_min:
            value = old_min
        if value > old_max:
            value = old_max
    return ((value - old_min) / (old_max - old_min)) * (new_max - new_min) + new_min


def chunks(item_list: ty.Iterable[ty.Any], n_items: int = 0, n_tasks: int = 0) -> ty.Generator:
    """Yield successive n-sized chunks from `item_list`.

    Parameters
    ----------
    item_list : list
        list of items
    n_items : int
        number of items in each chunk
    n_tasks : int
        number of tasks to be performed

    Returns
    -------
    chunk : list
        list of items
    """
    if n_items == 0 and n_tasks == 0:
        raise ValueError("You must specified either 'n_items' or 'n_tasks'.")
    if n_tasks:
        n_items = ceil(len(item_list) / n_tasks)
    for i in range(0, len(item_list), n_items):
        yield item_list[i : i + n_items]


def random_chunks(item_list: ty.Iterable[ty.Any], n_items: int = 0, n_tasks: int = 0) -> ty.Generator:
    """Randomly yield successive n-sized chunks from `item_list`.

    Warning: This function will shuffle the list ahead of time.
    """
    if n_items == 0 and n_tasks == 0:
        raise ValueError("You must specify either 'n_items' or 'n_tasks'. Both are zero")
    if n_tasks:
        n_items = ceil(len(item_list) / n_tasks)
    shuffle(item_list)
    for i in range(0, len(item_list), n_items):
        yield item_list[i : i + n_items]


def zip_chunks(item_list, n_items: int = 0, n_tasks: int = 0, *items):
    """Yield successive n-sized chunks from `item_list`."""
    yield from zip(
        chunks(item_list, n_items=n_items, n_tasks=n_tasks),
        *[chunks(x, n_items=n_items, n_tasks=n_tasks) for x in items],
    )


def sequential_chunks(item_list, n_items: int):
    """Create multiple lists of chunks that will be subsequent in nature.

    For instance if the `item_list` has
    the following values: [0, 1, 2, 3] and `n_items=2`, it will generate two lists of [0, 2] and [1, 3]. This can
    be helpful when doing tasks in chunks but want to see results of subsequent chunks.
    """
    n = len(item_list)
    n_pots = ceil(n / n_items)
    pots = [[] for _ in range(n_pots)]
    pot, current = 0, 0
    while current < n:
        pots[pot].append(item_list[current])
        pot += 1
        if pot >= n_pots:
            pot = 0
        current += 1
    yield from pots


def get_min_max(values: np.ndarray | ty.Iterable | list) -> tuple[int | float, int | float]:
    """Get the minimum and maximum value of an array."""
    return np.min(values), np.max(values)


def need_rotation(array: np.ndarray | tuple[int, ...]) -> bool:
    """Check whether image needs to be rotated."""
    shape = array.shape if not isinstance(array, tuple) else array
    if len(shape) == 3:
        return shape[1] > shape[2]
    return shape[0] > shape[1]


def rotate(array: np.ndarray, auto_rotate: bool) -> np.ndarray:
    """Rotate but only if user requested rotation."""
    return array if not auto_rotate else check_image_orientation(array)


def check_image_orientation(array: np.ndarray) -> np.ndarray:
    """Transpose image if the primary size is larger than the secondary size in order to improve images.

    Parameters
    ----------
    array: np.array
        2D heatmap

    Returns
    -------
    zvals: np.array
        (potentially) transposed 2D heatmap
    """
    shape = array.shape
    if len(shape) == 3:
        return np.swapaxes(array, 1, 2) if shape[1] > shape[2] else array
    return array.T if shape[0] > shape[1] else array
    # return np.rot90(array) if shape[0] > shape[1] else array


def slugify(value: ty.Any, allow_unicode: bool = False) -> str:
    """Convert to ASCII if 'allow_unicode' is False.

    Convert spaces or repeated dashes to single dashes. Remove characters that aren't alphanumerics, underscores,
    or hyphens. Convert to lowercase. Also strip leading and trailing whitespace, dashes, and underscores.
    """
    import unicodedata

    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize("NFKC", value)
    else:
        value = unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    value = re.sub(r"[^.=\w\s-]", "", value.lower())
    return re.sub(r"[-\s]+", "-", value).strip("-_")


def get_array_mask(array: np.ndarray, min_val: float, max_val: float):
    """Return mask for array."""
    return np.logical_and(array >= min_val, array <= max_val)


@nb.njit(parallel=True)
def get_array_mask_(array: np.ndarray, min_val: int | float, max_val: int | float) -> np.ndarray:
    """Return mask for array."""
    mask = np.empty(array.shape, dtype=np.bool_)
    for i in nb.prange(array.size):
        val = array[i]
        mask[i] = min_val <= val <= max_val
    return mask


@nb.njit(parallel=True)
def filter_array(
    array: np.ndarray, array_to_mask: np.ndarray, min_val: int | float, max_val: int | float
) -> np.ndarray:
    """Filter array based on another array."""
    # First, count how many elements pass the condition
    count = 0
    for i in nb.prange(array_to_mask.size):
        val = array_to_mask[i]
        if min_val <= val <= max_val:
            count += 1

    # Allocate an output array of the required size
    result = np.zeros(count, dtype=array.dtype)

    # Second pass: fill the result with filtered values
    idx = 0
    for i in nb.prange(array_to_mask.size):
        val = array_to_mask[i]
        if min_val <= val <= max_val:
            result[idx] = array[i]
            idx += 1
    return result


def get_array_window(array: np.ndarray, min_val: float, max_val: float, *arrays: np.ndarray) -> list[np.ndarray]:
    """Get narrower view of array based on upper and lower limits.

    The first array is the one that is used to create mask.
    """
    mask = np.logical_and(array >= min_val, array <= max_val)
    _arrays = [array[mask]]
    for _array in arrays:
        if _array.shape[0] != mask.shape[0]:
            raise ValueError("Incorrect shape of the input arrays")
        _arrays.append(_array[mask])
    return _arrays


def split_array(array: np.ndarray, chunk_size: int, axis=0):
    """Split array according to chunk size."""
    n = array.shape[axis]
    while True:
        if n <= chunk_size:
            yield array
            break
        else:
            if axis == 0:
                yield array.iloc[:chunk_size]
                array = array.iloc[chunk_size:]
            else:
                yield array[:, :chunk_size]
                array = array[:, chunk_size:]
            n -= chunk_size


def _remove_duplicates_from_dict(data):
    """Remove duplicates from list of dictionaries."""
    # list of dictionaries
    if isinstance(data, list):
        # check if list of dictionaries
        if all(isinstance(d, dict) for d in data):
            return [dict(t) for t in {tuple(d.items()) for d in data}]
    return data


class Cycler:
    """Cycling class similar to itertools.cycle with the addition of `previous` functionality."""

    def __init__(self, c):
        self._c = c
        self._index = -1

    def __len__(self):
        return len(self._c)

    def __next__(self):
        self._index += 1
        if self._index >= len(self._c):
            self._index = 0
        return self._c[self._index]

    def next(self):
        """Go forward."""
        return self.__next__()

    def previous(self):
        """Go backwards."""
        self._index -= 1
        if self._index < 0:
            self._index = len(self._c) - 1
        return self._c[self._index]

    def current(self):
        """Get current index."""
        return self._c[self._index]

    def set_current(self, index: int):
        """Set current index."""
        self._index = index


def difference_matrix(a: np.ndarray) -> np.ndarray:
    """Difference matrix.

    Compute difference between each element in a 1d array.
    """
    a = np.asarray(a)
    assert a.ndim == 1, "Input must be 1d array."

    x = np.reshape(a, (len(a), 1))
    return x - x.transpose()



def get_close_matches_case(word, possibilities, *args: ty.Any, **kwargs: ty.Any):
    """Case-insensitive version of difflib.get_close_matches."""
    from difflib import get_close_matches

    lower_word = word.lower()
    lower_possibilities = {p.lower(): p for p in possibilities}
    lower_matches = get_close_matches(lower_word, lower_possibilities.keys(), *args, **kwargs)
    return [lower_possibilities[m] for m in lower_matches]


def get_close_matches_case_msg(word: str, possibilities: ty.Iterable[str]) -> str:
    """Get message for close matches."""
    match = get_close_matches_case(word, possibilities)
    match = f"Did you mean: `{', '.join(match)}`?" if match else ""
    return f"Unknown parameter: `{word}`.{match}"


def calculate_array_size(a, as_str: bool = True):
    """Calculate the size of an array in Mb."""
    if not hasattr(a, "nbytes"):
        try:
            n_bytes = a.data.nbytes + a.indptr_col.nbytes + a.indices_row.nbytes
        except AttributeError:
            try:
                n_bytes = a.col.nbytes + a.row.nbytes + a.data.nbytes
            except AttributeError:
                n_bytes = a.data.nbytes + a.indptr.nbytes + a.indices.nbytes

    else:
        n_bytes = a.nbytes
    if as_str:
        return format_size(n_bytes)
    return n_bytes


def calculate_quantile_without_zeros(data, q=0.99):
    """Calculate quantile value of a heatmap.

    Parameters
    ----------
    data : np.array
        heatmap
    q : float, optional
        quantile value, by default 0.99

    Returns
    -------
    q_value : float
        intensity value at qth quantile
    """
    data = data.astype(np.float32)
    data[data == 0] = np.nan
    return np.nanquantile(data, q)


def get_distributed_list(
    total: int,
    n_frames_or_proportion: int | float,
    framelist: np.ndarray | None = None,
    as_index: bool = False,
    trim: bool = False,
    offset: int = 0,
) -> np.ndarray:
    """Get list of frames that span the entire dataset.

    Parameters
    ----------
    total: int
        total number of frames available
    n_frames_or_proportion : int / float
        number of frames or proportion of frames
    framelist : Optional[List, np.ndarray]
        list of frames to use for sub-sampling
    as_index : bool, optional
        if `True`, the distributed framelist will consist of indices rather than actual frame IDs
    trim : bool, optional
        if `True`, the distributed framelist will be trimmed to the maximum available frame
    offset : int
        Offset  to apply to the start of the frame list.

    Returns
    -------
    np.array
        array with list of frames to extract from the dataset
    """
    start = 0 if as_index else 1
    start += offset
    end = total

    # check whether n_frames is not a fraction
    if n_frames_or_proportion <= 1.0:
        n_frames_or_proportion = total * n_frames_or_proportion

    # make sure the number of requested frames does not exceed the maximum available frames
    if n_frames_or_proportion > end:
        n_frames_or_proportion = end

    # compute divider
    divider = floor(end / n_frames_or_proportion)

    # get frame list
    distributed = np.arange(start, end, divider).astype(np.int32)

    # frame_list above is simply index positions
    if framelist is not None:
        framelist = np.asarray(framelist)
        distributed = framelist[distributed]
    if trim and len(distributed) > n_frames_or_proportion:
        distributed = distributed[:n_frames_or_proportion]
    return distributed


def optimize_dtype(int_value=None, float_value=None, allow_unsigned: bool = False, guess_type: bool = True):
    """Compute the most optimal dtype for integer/float value.

    Parameters
    ----------
    int_value : int, optional
        integer value to test
    float_value : float, optional
        float value to test
    allow_unsigned : bool
        if 'True' unsigned version of the integer dtype will be used
    guess_type : bool
        if 'True', value will be checked to be either integer or float

    Returns
    -------
    dtype : np.dtype
        optimal dtype for particular value
    """

    def _check_dtype(value):
        if np.issubdtype(type(value), np.integer):
            _int_value, _float_value = value, None
        elif np.issubdtype(type(value), np.floating):
            _int_value, _float_value = None, value
        else:
            raise ValueError(f"Could not determine dtype - {type(value)}")
        return _int_value, _float_value

    if all(value is not None for value in [int_value, float_value]):
        raise ValueError("You should either specify 'int_value' or 'float_value' and not both")

    if all(value is None for value in [int_value, float_value]):
        raise ValueError("You should either specify 'int_value' or 'float_value' and not None")

    # guess data type based on the input
    if guess_type:
        int_value, float_value = _check_dtype(int_value if int_value is not None else float_value)

    if int_value is not None:
        if int_value < 0:
            allow_unsigned = False

        int8 = np.uint8 if allow_unsigned else np.int8
        int16 = np.uint16 if allow_unsigned else np.int16
        int32 = np.uint32 if allow_unsigned else np.int32
        int64 = np.uint64 if allow_unsigned else np.int64
        if np.iinfo(int8).min <= int_value <= np.iinfo(int8).max:
            return int8
        if np.iinfo(int16).min <= int_value <= np.iinfo(int16).max:
            return int16
        if np.iinfo(int32).min <= int_value <= np.iinfo(int32).max:
            return int32
        return int64
    if float_value is not None:
        if np.finfo(np.float32).min <= float_value <= np.finfo(np.float32).max:
            return np.float32
        return np.float64


def find_nearest_divisible(value: int | float, divisor: int | float, max_iters: int = 1000) -> int | float:
    """Find nearest value that can be evenly divided by the divisor.

    Parameters
    ----------
    value : Union[int, float]
        value to be divided by the divisor
    divisor : Union[int, float]
        value by which the `value` is divided
    max_iters : int
        maximum number of iterations before the algorithm should give up

    Returns
    -------
    value : Union[int, float]
        new value if the algorithm did not fail to find new value or -1 if it did
    """
    n_iter = 0
    while value % divisor != 0 and n_iter < max_iters:
        value += 1
        n_iter += 1

    if value % divisor == 0:
        return value
    return -1


def view_as_blocks(array: np.ndarray, n_rows: int, n_cols: int, auto_pad: bool = True):
    """Return an array of shape (n, n_rows, n_cols) where n * n_rows * n_cols = array.size.

    If array is a 2D array, the returned array should look like n sub-blocks with
    each sub-block preserving the "physical" layout of array.


    Parameters
    ----------
    array : np.ndarray
        input array
    n_rows : int
        number of rows in each sub-block
    n_cols : int
        number of columns in each sub-block
    auto_pad : bool, optional
        if `True`, the array will be padded with NaNs to ensure the input array is actually divisible by `n_rows` and
        `n_cols`

    References
    ----------
    Inspired by a StackOverflow post [1]
    [1] https://stackoverflow.com/questions/16856788/slice-2d-array-into-smaller-2d-arrays
    """
    array = np.asarray(array)

    h, w = array.shape
    if auto_pad:
        # if the shape is incorrect, pad the array with NaNs at the outer edges
        if h % n_rows != 0 or w % n_cols != 0:
            new_height = find_nearest_divisible(h, n_rows)
            new_width = find_nearest_divisible(w, n_cols)
            _array = array
            array = np.full((new_height, new_width), fill_value=np.nan)
            array[:h, :w] = _array
            h, w = array.shape

    # check shape
    if h % n_rows != 0:
        raise ValueError(
            f"{h} rows is not evenly divisible by {n_rows}. Nearest alternative: {find_nearest_divisor(h, n_rows)}"
        )
    if w % n_cols != 0:
        raise ValueError(
            f"{w} cols is not evenly divisible by {n_cols}. Nearest alternative: {find_nearest_divisor(w, n_cols)}"
        )
    new_shape = (int(h / n_rows), int(w / n_cols))
    return array.reshape((h // n_rows, n_rows, -1, n_cols)).swapaxes(1, 2).reshape(-1, n_rows, n_cols), new_shape


def prettify_name(mz_min: float, mz_max: float, prefix: str = "", suffix: str = "", n_decimals: int = 0) -> str:
    """Prettify ion name."""
    mz_window = (mz_max - mz_min) / 2
    mz_val = mz_min + mz_window  # estimated value
    return f"{prefix}{mz_val:.3f} ± {round(mz_window, n_decimals)}{suffix}"


def prettify_ion_name(mz_min: float, mz_max: float):
    """Prettify name."""
    return prettify_name(mz_min, mz_max, suffix=" Da", n_decimals=3)



"""Timer utilities."""

from __future__ import annotations

import time
import typing as ty
from contextlib import contextmanager
from datetime import datetime, timedelta

PERIODS = [
    ("year", 60 * 60 * 24 * 365 * 1e9),
    ("month", 60 * 60 * 24 * 30 * 1e9),
    ("day", 60 * 60 * 24 * 1e9),
    ("h", 60 * 60 * 1e9),
    ("min", 60 * 1e9),
    ("s", 1 * 1e9),
    ("ms", 1 * 1e6),
    ("μs", 1 * 1e3),
    ("ns", 1),
]


def format_datetime_ago(datetime_str: str) -> str:
    """
    Format a datetime string into a human-readable 'time ago' format.

    Parameters
    ----------
    datetime_str : str
        Datetime string in the format 'YYYY-MM-DDTHH:MM:SSZ'.

    Returns
    -------
    str
        Human-readable 'time ago' format.
    """
    if not datetime_str:
        return ""

    try:
        # Parse the input datetime string
        datetime_obj = datetime.strptime(datetime_str, "%Y-%m-%dT%H:%M:%SZ")

        # Calculate the time difference from the current time
        now = datetime.utcnow()
        time_difference = now - datetime_obj

        # Define time units
        units = [
            ("year", timedelta(days=365)),
            ("month", timedelta(days=30)),
            ("day", timedelta(days=1)),
            ("hour", timedelta(hours=1)),
            ("minute", timedelta(minutes=1)),
        ]

        # Find the largest time unit that fits
        for unit, delta in units:
            if time_difference >= delta:
                time_ago = time_difference // delta
                if time_ago > 1:
                    unit += "s"
                return f"{time_ago} {unit} ago"

        # If less than a minute, show 'just now'
        return "just now"

    except ValueError:
        return "Invalid datetime format"


def format_human_time_s(seconds: float, n_max: int = 2) -> str:
    """Format time so its human-readable."""
    return format_human_time(seconds * 1e9, n_max)


def format_human_time(nanoseconds: float, n_max: int = 2) -> str:
    """Format time so its human-readable."""
    values = []
    for period_name, period_seconds in PERIODS:
        if nanoseconds > period_seconds:
            period_value, nanoseconds = divmod(nanoseconds, period_seconds)
            values.append(f"{int(period_value)}{period_name}")
    if len(values) > n_max:
        values = values[0:n_max]
    return " ".join(values)


def format_time(seconds: float) -> str:
    """Convert time to nicer format. Value is provided in seconds."""
    if seconds <= 0.01:
        return f"{seconds * 1000000:.0f}us"
    elif seconds <= 0.1:
        return f"{seconds * 1000:.1f}ms"
    elif seconds > 86400:
        return f"{seconds / 86400:.2f}day"
    elif seconds > 1800:
        return f"{seconds / 3600:.2f}hr"
    elif seconds > 60:
        return f"{seconds / 60:.2f}min"
    return f"{seconds:.2f}s"


class MeasureTimer:
    """Timer class."""

    def __init__(self, func: ty.Callable | None = None, msg: str | None = None, human: bool = True):
        self.func = func
        self.msg = msg
        self.human = human
        if self.func and not self.msg:
            self.msg = "Execution took: {}"
        if self.msg and "{}" not in self.msg:
            self.msg += " in {}"

        self.start: float = float(time.perf_counter_ns() if human else time.perf_counter())
        self.end: float | None = None
        self.last: float | None = None
        self.steps: list[str] = []

    def stopwatch(self, name: str = "") -> None:
        """Add a step to the timer."""
        self.steps.append(self(since_last=True, name=name, with_steps=False))

    def finish(self) -> MeasureTimer:
        """Finish the timer."""
        self.end = self.current()
        return self

    def current(self) -> float:
        """Return current time."""
        return time.perf_counter_ns() if self.human else time.perf_counter()

    def elapsed(self, n: int = 1, start: float | None = None) -> float:
        """Return amount of time that elapsed."""
        end = self.end or self.current()
        start = start or self.start
        elapsed = end - start
        elapsed = elapsed / n
        self.last = end
        return elapsed

    def elapsed_since_last(self) -> float:
        """Elapsed since last time."""
        return self.elapsed(start=self.last)

    def format(self, elapsed: int) -> str:
        """Format time."""
        return format_human_time(elapsed) if self.human else format_time(elapsed)

    def __call__(
        self,
        n: int = 1,
        current: int = 1,
        start: float | None = None,
        since_last: bool = False,
        with_steps: bool = True,
        name: str = "",
    ) -> str:
        if since_last:
            start = self.last
        elapsed = self.elapsed(n, start)
        formatted = format_human_time(elapsed) if self.human else format_time(elapsed)
        if n > 1:
            formatted = f"{formatted} [{current}/{n}]"
        if name:
            formatted = f"{formatted} ({name})"
        if with_steps and self.steps:
            formatted += f" | {', '.join(self.steps)}"
        return formatted

    def __repr__(self) -> str:
        """Return nicely formatted execution time."""
        return self()

    def __str__(self) -> str:
        """Return nicely formatted execution time."""
        return self()

    def __enter__(self) -> MeasureTimer:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.finish()
        if self.func and self.msg:
            self.func(self.msg.format(self()))


@contextmanager
def measure_time():
    """Function that measures how long a function took place.

    Examples
    --------
    func = lambda: time.sleep(1)
    with measure_time() as t:
        func()
    t()
    """
    start = time.perf_counter()
    yield lambda: time.perf_counter() - start


@contextmanager
def report_measure_time(human: bool = True) -> ty.Generator[ty.Callable, None, None]:
    """Report measured time and print it."""
    if human:
        start = time.perf_counter_ns()
        yield lambda: format_human_time(time.perf_counter_ns() - start)
    else:
        start = time.perf_counter()
        yield lambda: format_time(time.perf_counter() - start)


def report_time(t_start: float):
    """Returns nicely formatted execution time."""
    return format_time(time.time() - t_start)


def time_loop(t_start: float, n_item: int, n_total: int, as_percentage: bool = True) -> str:
    """Calculate average, remaining and total times.

    Parameters
    ----------
    t_start : float
        starting time of the for loop
    n_item : int
        index of the current item - assumes index starts at 0
    n_total : int
        total number of items in the for loop - assumes index starts at 0
    as_percentage : bool, optional
        if 'True', progress will be displayed as percentage rather than the raw value

    Returns
    -------
    timed : str
        loop timing information
    """
    t_tot = time.time() - t_start
    t_avg = t_tot / (n_item + 1)
    t_rem = t_avg * (n_total - n_item + 1)

    # calculate progress
    progress = f"{n_item}/{n_total + 1}"
    if as_percentage:
        progress = f"{(n_item / (n_total + 1)) * 100:.1f}% ({progress})"

    return f"[Avg: {format_time(t_avg)} | Rem: {format_time(t_rem)} | Tot: {format_time(t_tot)} || {progress}]"


def time_average(t_start: float, n_total: int) -> str:
    """Calculate average and total time of a task.

    Parameters
    ----------
    t_start : float
        starting time of the task
    n_total : int
        total number of items

    Returns
    -------
    value : str
        formatted text
    """
    t_tot = time.time() - t_start
    t_avg = t_tot / (n_total + 1)

    return f"[Avg: {format_time(t_avg)} | Tot: {format_time(t_tot)}]"


class Timer:
    """Timer class."""

    def __init__(self, value=None, init=False):
        """Initialize timer."""
        self.timers = {}
        self._last_key = None
        if value:
            self.append(value)

        if init:
            self.qappend()

    def reset(self):
        """Reset timer."""
        self.timers = {}

    def append(self, value, label=""):
        """Adds new timed object to the dict."""
        _n = len(self.timers)
        if label == "":
            label = f"timer {_n + 1}"
        self.timers[_n] = [value, label]
        self._last_key = _n

    def qappend(self, label=""):
        self.append(time.time(), label)

    def get(self, add_total=True, sep="\n"):
        """Return computed timings."""
        keys = list(self.timers.keys())
        n_keys = len(keys)

        if n_keys < 2:
            return

        # get total
        total = time.time() - self.timers[0][0]

        timings = []
        for key in reversed(keys):
            value_curr, label = self.timers[key]
            if key >= 1:
                value_prev, __ = self.timers[key - 1]
                value = value_curr - value_prev
                try:
                    value_percent = (value / total) * 100
                except ZeroDivisionError:
                    value_percent = 0
                timings.append(f"{label}: {format_time(value)} ({value_percent:.2f}%)")

        # ensures nicer formatting
        timings.append("")

        # add total
        if add_total:
            timings.insert(0, f"TOTAL: {total:.4f}s")

        return sep.join(reversed(timings))

    def show(self, add_total=True, sep="\n"):
        """Returns computed timings."""
        self.get(add_total, sep)

    def last(self):
        """Retrieve last time."""
        return self.timers[self._last_key][0]

    def first(self):
        """Retrieve first time."""
        return self.timers[0][0]



"""Multi-core utilities."""

import math
import time
import typing as ty
import warnings

from joblib import Parallel as _Parallel
from loguru import logger
from psutil import cpu_count, virtual_memory
from tqdm import tqdm

from koyo.timer import report_time
from koyo.utilities import running_as_pyinstaller_app


def estimate_cpu_count_from_size(max_obj_size_in_bytes, keep_free_in_bytes=4_000_000):
    """Get number of cores based on the amount of RAM and maximum size of an object.

    Parameters
    ----------
    max_obj_size_in_bytes : int
        maximum size of an object in bytes. If specified in GB, it will be automatically converted to bytes
    keep_free_in_bytes : int, optional
        minimal amount of RAM reserved for the OS and other actions, by default 4000000
        If specified in Gb, it will be automatically converted to bytes

    Returns
    -------
    n_cores : int
        number of cores to be used by any multicore enabled function

    Raises
    ------
    ValueError
        raises ValueError if number of cores is less than 1
    """
    # convert to GB
    if int(math.log10(keep_free_in_bytes)) + 1 <= 2:
        keep_free_in_bytes = keep_free_in_bytes * 1024**3
    if int(math.log10(max_obj_size_in_bytes)) + 1 <= 2:
        max_obj_size_in_bytes = max_obj_size_in_bytes * 1024**3

    # get currently available memory
    available_memory = virtual_memory().total - keep_free_in_bytes
    n_cores = round(available_memory / max_obj_size_in_bytes)
    if n_cores < 1:
        raise ValueError(
            "Based on the amount of RAM available on this system, there is not enough memory to perform"
            + " this action."
        )

    if n_cores > get_cpu_count():
        n_cores = get_cpu_count()
    print(
        f"Total RAM: {virtual_memory().total / 1024**3:0f} Gb",
        f"\nReserved memory: {keep_free_in_bytes / 1024**3} Gb",
        f"\nMax. size of object: {max_obj_size_in_bytes / 1024**3:0f} Gb",
        f"\nMax. recommended cores: {n_cores}",
        sep="",
    )
    return n_cores


def get_cpu_count(keep_free: int = 2, max_cpu: int = 24, n_tasks: int = 0) -> int:
    """Get number of cores available on the machine.

    Parameters
    ----------
    keep_free : int, optional
        using all cores can significantly slow the system down, so we allow for some to remain free, by default 2
    max_cpu : int
        maximum number of cores
    n_tasks : int
        number of tasks to be executed

    Returns
    -------
    n_cores : int
        number of cores to be used by any multicore enabled function

    Raises
    ------
    ValueError
        raises ValueError if number of cores is less than 1
    """
    n_cores = cpu_count() - keep_free
    if n_cores > max_cpu:
        n_cores = max_cpu
    if n_tasks > 0 and n_cores > n_tasks:
        n_cores = n_tasks
    if n_cores < 1:
        warnings.warn(
            f"Tried to reserve `{keep_free}` cores free ut failed be cores. Action will use all cores.",
            RuntimeWarning,
            stacklevel=2,
        )
        n_cores = cpu_count()
    return n_cores


def is_daemon() -> bool:
    """Check if current process is daemonic."""
    from multiprocessing import current_process

    return current_process().daemon


class MultiCoreExecutor:
    """MPIRE-based executor."""

    def __init__(self, n_cores: int, silent: bool = False, desc=""):
        self.n_cores = n_cores
        self.silent = silent
        self.desc = desc

    def run(self, func: ty.Callable, args: ty.Iterable, auto_expand: bool = True):
        """Execute."""
        import mpire

        kws = {
            "worker_lifespan": 1,
            "progress_bar": not self.silent,
            "progress_bar_options": {"desc": self.desc, "mininterval": 5},
        }

        res = []
        if auto_expand:
            if args:
                args = list(args)
                if len(args) == 1 or self.n_cores == 1:
                    for arg in tqdm(args, disable=False, desc=self.desc):
                        res.append(func(*arg))
                else:
                    with mpire.WorkerPool(n_jobs=self.n_cores, keep_alive=False) as pool:
                        res = pool.map(func, args, iterable_len=len(args), **kws)
        else:
            with mpire.WorkerPool(n_jobs=self.n_cores, keep_alive=False) as pool:
                res = pool.map(func, args, **kws)
        return res

    def lazy_run(self, func: ty.Callable, args: ty.Iterable, auto_expand: bool = True):
        """Execute."""
        import mpire

        kws = {
            "worker_lifespan": 1,
            "progress_bar": not self.silent,
            "progress_bar_options": {"desc": self.desc, "mininterval": 5},
        }

        res = []
        if auto_expand:
            if args:
                args = list(args)
                if len(args) == 1 or self.n_cores == 1:
                    for arg in tqdm(args, disable=False, desc=self.desc):
                        res.append(func(*arg))
                        yield res[-1]
                else:
                    with mpire.WorkerPool(n_jobs=self.n_cores, keep_alive=True) as pool:
                        for res_ in pool.imap(func, args, iterable_len=len(args), **kws):
                            res.append(res_)
                            yield res_
        else:
            with mpire.WorkerPool(n_jobs=self.n_cores, keep_alive=True) as pool:
                for res_ in pool.imap(func, args, **kws):
                    res.append(res_)
                    yield res_
        return res


class MultiCoreDispatcher:
    """Base dispatcher class."""

    def __init__(self, n_cores: int = -1, chunksize: int = 100, backend: str = "mpire", quiet: bool = False, **kwargs):
        # accessible attributes
        self.n_cores = self.optimize_n_cores(n_cores, array_size=kwargs.get("array_size"))
        self.chunksize = self.optimize_chunksize(chunksize, self.n_cores)
        self.backend = backend if not running_as_pyinstaller_app() else "multiprocessing"

        # private attributes
        self._maxtasksperchild = kwargs.get("maxtasksperchild", 1)
        self.quiet = quiet

        logger.info(f"Using {self.n_cores} cores. Chunksize {self.chunksize}. Backend: {self.backend}")

    @staticmethod
    def optimize_n_cores(n_cores: int, array_size=None):
        """Calculate the optimal number of cores based on some criteria."""
        # check whether we should optimize it in first place
        if n_cores in [-1, 0, 512]:
            if array_size is not None:
                return estimate_cpu_count_from_size(array_size, 30)
            return get_cpu_count(4)
        if not isinstance(n_cores, int):
            raise ValueError("The value of 'n_cores' must be an integer!")
        return n_cores

    @staticmethod
    def optimize_n_threads(n_threads: int, n_cores: int):
        """If the number of cores exceeds 1, it is usually best not to do multi-threading."""
        if n_cores > 1:
            return 1
        if isinstance(n_threads, int):
            return n_threads
        return 1

    @staticmethod
    def optimize_chunksize(chunksize, n_cores):
        """Optimize chunksize."""
        # if chunk size is set to -1, set it to maximise speed by using maximum number of cores
        # if chunksize == -1:
        #     chunksize = np.max([math.ceil(n_extraction_windows / self._n_cores), 100])
        return chunksize

    def generate_iterable(self) -> ty.List:
        """Return iterable which can be consumed by the `generate_args_list` function."""
        return []

    def generate_args_list(self, post_process: bool = False):
        """Generate list of arguments which can be handled by the `dispatch` function."""
        raise NotImplementedError("Method must be implemented by the subclass")

    def dispatch(self, func: ty.Callable, *, pre_process: bool = True, post_process: bool = True):
        """Dispatches the job over multiple cores to speed-up processing.

        Parameters
        ----------
        func : Callable
            callable function which handles the processing
        pre_process : bool
            flag to run pre-processing before starting tasks
        post_process : bool
            flag to run post-processing after completing all tasks

        Notes
        -----
        Arguments have to pickleable
        """
        t_start = time.time()
        if pre_process:
            self.pre_process()

        try:
            args_list = self.generate_args_list(post_process)
        except TypeError:
            args_list = self.generate_args_list()
        if len(args_list) < self.n_cores:
            self.n_cores = len(args_list)
        if args_list:
            logger.info(f"{len(args_list)} tasks in the queue...")
            if self.n_cores == 1:
                self._single_executor(func, args_list)
            elif self.backend == "joblib":
                self._joblib_executor(func, args_list)
            else:
                self._mpire_executor(func, args_list)

        else:
            logger.info("There are no jobs to be done.")

        # post-process
        if post_process and args_list and self.n_cores > 0:
            self.post_process()
        logger.info(f"*** COMPLETED ACTION IN {report_time(t_start)} ***")

    def _single_executor(self, func, args_list):
        """Single-core executor."""
        for args in tqdm(args_list, disable=self.quiet):
            func(*args)

    def _joblib_executor(self, func, args_list):
        """Joblib executor."""
        from joblib import delayed

        _args_list = []
        for arg in args_list:
            _args_list.append(delayed(func)(*arg))
        del args_list
        Parallel(
            n_jobs=self.n_cores,
            verbose=50,
            prefer="processes",
        )(_args_list)

    def _multiprocessing_executor(self, func, args_list):
        """Multiprocessing executor."""
        import multiprocessing

        if args_list:
            # create processor pool
            with multiprocessing.Pool(processes=self.n_cores, maxtasksperchild=self._maxtasksperchild) as pool:
                # map function with tuple of arguments
                pool.starmap(func, args_list)
            # close pool and join it until all jobs have finished
            pool.close()
            pool.join()

    def _mpire_executor(self, func, args_list):
        """MPIRE executor."""
        import mpire

        if args_list:
            # create processor pool
            with mpire.WorkerPool(n_jobs=self.n_cores) as pool:
                pool.map_unordered(
                    func,
                    args_list,
                    iterable_len=len(args_list),
                    worker_lifespan=self._maxtasksperchild,
                    progress_bar=not self.quiet,
                )

    def pre_process(self):
        """Pre-process."""

    def post_process(self):
        """Post-process."""


class Parallel(_Parallel):
    """Parallel."""

    def __init__(self, *args, **kwargs):
        if "backend" in kwargs and kwargs["backed"] == "loky" and running_as_pyinstaller_app():
            kwargs["backend"] = "multiprocessing"
        if "backend" not in kwargs and running_as_pyinstaller_app():
            kwargs["backend"] = "multiprocessing"
        logger.info(f"Executing using backend - {kwargs.get('backend', 'loky')}")
        super().__init__(*args, **kwargs)


class ProgressParallel(Parallel):
    """joblib's Parallel wrapper that includes tqdm progress indicator."""

    def __init__(self, disable=False, total=None, desc: str = "", *args, **kwargs):
        self._disable = disable
        self._total = total
        self._desc = desc
        super().__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        with tqdm(disable=self._disable, total=self._total, desc=self._desc) as self.progress_bar:
            return Parallel.__call__(self, *args, **kwargs)

    def print_progress(self):
        """Print progress."""
        if self._total is None:
            self.progress_bar.total = self.n_dispatched_tasks
        self.progress_bar.n = self.n_completed_tasks
        self.progress_bar.refresh()



"""Plotly functions."""

from __future__ import annotations

import typing as ty
from pathlib import Path

from koyo.typing import PathLike
from koyo.utilities import is_installed

if ty.TYPE_CHECKING:
    try:
        from plotly.graph_objects import Figure
    except ImportError:
        Figure = None


def write_html(fig: Figure, filename: PathLike) -> Path:
    """Write HTML to file."""
    filename = Path(filename)
    filename.parent.mkdir(parents=True, exist_ok=True)
    fig.write_html(str(filename), include_plotlyjs="cdn")
    return filename


def write_png(fig: Figure, filename: PathLike) -> Path | None:
    """Write PNG."""
    filename = Path(filename).with_suffix(".png")
    if is_installed("kaleido"):
        filename.parent.mkdir(parents=True, exist_ok=True)
        fig.write_image(str(filename))
        return filename


def show_html(filename: PathLike) -> None:
    """Show HTML in browser."""
    import webbrowser

    filename = Path(filename)
    webbrowser.open(filename.as_uri())



"""Test PDF utilities."""

from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from koyo.pdf_mixin import PDFMixin
from PIL import Image


class PDF(PDFMixin):
    """Mixin class."""

    def __init__(self, filename: Path, as_pdf: bool):
        self._filename = filename
        self.as_pdf = as_pdf

    @property
    def pdf_filename(self) -> Path:
        """Filename."""
        return self._filename


def test_pdf(tmp_path):
    tmp = Path(tmp_path)
    # init PDF
    pdf = PDF(tmp / "test.pdf", as_pdf=True)

    assert pdf.pdf_filename == tmp / "test.pdf"
    array = np.random.randint(0, 255, (1000, 1000), dtype=np.uint8)
    image = Image.fromarray(array)
    pdf._add_pil_image_to_pdf(tmp / "test.png", image)
    pdf._save_pdf(pdf.pdf)
    assert (tmp / "test.pdf").exists()

    # create random page
    page = pdf._make_pdf(tmp / "test2.pdf")
    assert page is not None
    # export title
    pdf._add_title_to_pdf("Title", page)
    # export figure
    plt.plot([1, 2, 3], [1, 2, 3])
    pdf._add_mpl_figure_to_pdf(tmp / "test.png", plt.gcf(), pdf=page)
    assert not (tmp / "test.png").exists()
    # export PIL figure
    array = np.random.randint(0, 255, (1000, 1000), dtype=np.uint8)
    image = Image.fromarray(array)
    pdf._add_pil_image_to_pdf(tmp / "test.png", image, pdf=page)
    assert not (tmp / "test.png").exists()
    pdf._add_title_to_pdf("Title", page)
    pdf._save_pdf(page)
    assert (tmp / "test2.pdf").exists()

    with pdf._export_pdf_figures(tmp / "test3.pdf") as page:
        array = np.random.randint(0, 255, (1000, 1000), dtype=np.uint8)
        image = Image.fromarray(array)
        pdf._add_pil_image_to_pdf(tmp / "test.png", image, pdf=page)
    assert (tmp / "test3.pdf").exists()



"""Test pptx_mixin."""

from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pytest
from koyo.pptx_mixin import HAS_PPTX, PPTXMixin
from PIL import Image


class PPTX(PPTXMixin):
    """Mixin class."""

    def __init__(self, filename: Path, as_pptx: bool):
        self._filename = filename
        self.as_pptx = as_pptx

    @property
    def pptx_filename(self) -> Path:
        """Filename."""
        return self._filename


@pytest.mark.skipif(not HAS_PPTX, reason="pptx not installed")
def test_pptx(tmp_path):
    tmp = Path(tmp_path)
    # init PDF
    pptx = PPTX(tmp / "test.pptx", as_pptx=True)

    assert pptx.pptx_filename == tmp / "test.pptx"
    array = np.random.randint(0, 255, (1000, 1000), dtype=np.uint8)
    image = Image.fromarray(array)
    pptx._add_pil_image_to_pptx(tmp / "test.png", image, title="Title for plot")
    pptx._save_pptx(pptx.pptx)
    assert (tmp / "test.pptx").exists()

    # create random page
    page = pptx._make_pptx(tmp / "test2.pptx")
    assert page is not None
    # export figure
    plt.plot([1, 2, 3], [1, 2, 3])
    pptx._add_mpl_figure_to_pptx(tmp / "test.png", plt.gcf(), pptx=page)
    assert not (tmp / "test.png").exists()
    plt.plot([1, 2, 3], [1, 2, 3])
    pptx._add_mpl_figure_to_pptx(tmp / "test.png", plt.gcf(), pptx=page, title="Title for plot")
    assert not (tmp / "test.png").exists()
    # export PIL figure
    array = np.random.randint(0, 255, (1000, 1000), dtype=np.uint8)
    image = Image.fromarray(array)
    pptx._add_pil_image_to_pptx(tmp / "test.png", image, pptx=page, title="Title for plot")
    assert not (tmp / "test.png").exists()
    pptx._add_title_to_pptx("Title", page)
    pptx._add_content_to_pptx("Other text\nOther text\nOther text", "Title", page)
    pptx._save_pptx(page)
    assert (tmp / "test2.pptx").exists()

    with pptx._export_pptx_figures(tmp / "test3.pptx") as page:
        array = np.random.randint(0, 255, (1000, 1000), dtype=np.uint8)
        image = Image.fromarray(array)
        pptx._add_pil_image_to_pptx(tmp / "test.png", image, pptx=page, title="Title for plot")
    assert (tmp / "test3.pptx").exists()



"""Configuration module."""

import typing as ty
from contextlib import contextmanager, suppress
from pathlib import Path

from loguru import logger
from pydantic import BaseModel


class BaseConfig(BaseModel):
    """Base config."""

    USER_CONFIG_DIR: ty.ClassVar[Path]
    USER_CONFIG_FILENAME: ty.ClassVar[str] = "config.json"

    def __init__(self, _auto_load: bool = False, **kwargs: ty.Any):
        super().__init__(**kwargs)
        if _auto_load:
            self.load()

    @property
    def output_path(self) -> Path:
        """Get default output path."""
        self.USER_CONFIG_DIR.mkdir(parents=True, exist_ok=True)
        return self.USER_CONFIG_DIR / self.USER_CONFIG_FILENAME

    def get_exclude_fields(self) -> set[str]:
        """Get fields to exclude from saving."""
        exclude = []
        schema = self.model_json_schema()["properties"]
        for field_name, _field in self.model_fields.items():
            field_schema = schema[field_name]
            if not field_schema.get("save", True):
                exclude.append(field_name)
        return set(exclude)

    def _set_values(self, **kwargs: ty.Any) -> bool:
        changed = False
        for key, value in kwargs.items():
            if hasattr(self, key):
                old_value = getattr(self, key)
                try:
                    setattr(self, key, value)
                    if old_value != getattr(self, key):
                        changed = True
                except Exception as e:
                    logger.warning(f"Failed to set {key}={value}: {e}")
            else:
                logger.warning(f"[{self.__class__.__name__}] Unknown key {key}={value} - perhaps it was deprecated?")
        return changed

    def update(self, save: bool = True, **kwargs: ty.Any) -> None:
        """Update configuration and  pip install save to file."""
        changed = self._set_values(**kwargs)
        if save and changed:
            with suppress(OSError, PermissionError):
                self.save()

    def load(self) -> None:
        """Load configuration from file."""
        from koyo.json import read_json_data

        if self.output_path.exists():
            try:
                data = read_json_data(self.output_path)
                self._set_values(**data)
                logger.info(f"Loaded configuration from {self.output_path}")
            except Exception as e:
                logger.warning(f"Failed to load configuration from {self.output_path}: {e}")
                logger.exception(e)

    def save(self) -> None:
        """Export configuration to file."""
        try:
            self.output_path.write_text(
                self.model_dump_json(indent=4, exclude_unset=True, exclude=self.get_exclude_fields())
            )
            logger.info(f"Saved configuration to {self.output_path}")
        except Exception as e:
            logger.warning(f"Failed to save configuration to {self.output_path}: {e}")

    @contextmanager
    def temporary_overwrite(self, **kwargs: ty.Any) -> ty.Generator[None, None, None]:
        """Temporarily overwrite configuration and then revert back."""
        original = {key: getattr(self, key) for key in kwargs}
        for key, value in kwargs.items():
            setattr(self, key, value)
        yield
        for key, value in original.items():
            setattr(self, key, value)



"""Helper functions for PyInstaller hooks."""

from __future__ import annotations

import importlib.util
import sys
import types
from collections.abc import Iterator
from pathlib import Path


def load_module_from_path(path: Path, module_name: str | None = None) -> types.ModuleType:
    """
    Load a .py file from an arbitrary path and return the module object.

    - path: full path to the .py file (e.g. "PATH/TO/SCRIPT/hook-ionglow.py")
    - module_name: optional name to register in sys.modules (e.g. "hook_ionglow").
                   If None, a unique name based on file name will be used.
    """
    path = Path(path).expanduser().resolve()
    if not path.exists() or path.suffix != ".py":
        raise FileNotFoundError(f"{path} not found or not a .py file")

    # default module name: make filename a valid identifier
    if module_name is None:
        # convert "hook-ionglow.py" -> "hook_ionglow"
        module_name = path.stem.replace("-", "_")

    spec = importlib.util.spec_from_file_location(module_name, str(path))
    if spec is None or spec.loader is None:
        raise ImportError(f"Cannot create import spec for {path}")

    module = importlib.util.module_from_spec(spec)
    # register so imports inside the loaded module can use sys.modules if needed
    sys.modules[module_name] = module
    spec.loader.exec_module(module)  # runs the module code
    return module


def load_hooks(
    hook_dir: Path, pattern: str = "hook-*.py"
) -> tuple[list[str], list[tuple[str, str]], list[tuple[str, str]]]:
    """Load all hook modules in the given directory matching the pattern."""
    hiddenimports, datas, binaries = [], [], []
    for hook in hook_dir.glob(pattern):
        mod = load_module_from_path(hook)
        if hasattr(mod, "hiddenimports"):
            hiddenimports += mod.hiddenimports
        if hasattr(mod, "datas"):
            datas += mod.datas
        if hasattr(mod, "binaries"):
            binaries += mod.binaries
    return hiddenimports, datas, binaries


def get_runtime_hooks(hook_dir: Path, pattern: str = "hook-*.py") -> list[str]:
    """Get a list of runtime hooks."""
    return [str(hook) for hook in hook_dir.glob(pattern)]


def _iter_all_modules(
    package: str | types.ModuleType,
    prefix: str = "",
) -> Iterator[str]:
    """Iterate over the names of all modules that can be found in the given
    package, recursively.

        >>> import _pytest
        >>> list(_iter_all_modules(_pytest))
        ['_pytest._argcomplete', '_pytest._code.code', ...]
    """
    import os
    import pkgutil

    if isinstance(package, str):
        path = package
    else:
        # Type ignored because typeshed doesn't define ModuleType.__path__
        # (only defined on packages).
        package_path = package.__path__
        path, prefix = package_path[0], package.__name__ + "."
    for _, name, is_package in pkgutil.iter_modules([path]):
        if is_package:
            for m in _iter_all_modules(os.path.join(path, name), prefix=name + "."):
                yield prefix + m
        else:
            yield prefix + name



"""Fault handler."""

import atexit
from contextlib import suppress
from pathlib import Path

from loguru import logger

from koyo.typing import PathLike


def submit_sentry_attachment(message: str, path: PathLike) -> None:
    """Submit attachment to Sentry."""
    try:
        from sentry_sdk import configure_scope, get_current_scope

        scope = get_current_scope()
        scope.add_attachment(path=str(path))
        scope.capture_message(message)
    except ImportError:
        logger.exception("Failed to submit attachment to Sentry. Please report this issue to the developers.")
    except Exception:
        logger.exception("Failed to submit attachment to Sentry.")


def install_segfault_handler(output_dir: PathLike, filename: str = "segfault.log") -> None:
    """Install segfault handler."""
    import faulthandler

    segfault_path = Path(output_dir) / filename
    segfault_file = open(segfault_path, "w+")
    faulthandler.enable(segfault_file, all_threads=True)
    atexit.register(segfault_file.close)
    logger.trace(f"Enabled fault handler - logging to '{segfault_path}'")


# noinspection PyBroadException
def maybe_submit_segfault(output_dir: PathLike, filename: str = "segfault.log") -> None:
    """Submit segfault to Sentry if there is an existing segfault file."""
    from datetime import datetime

    segfault_path = Path(output_dir) / filename
    if not segfault_path.exists():
        return

    try:
        # read segfault data
        segfault_text = segfault_path.read_text()
        if not segfault_text:
            return

        logger.error("There was a segmentation fault previously - submitting to Sentry if it's available.")

        # create backup of the segfault file
        try:
            segfault_backup_path = Path(output_dir) / f"segfault_{datetime.now().strftime('%Y%m%d_%H%M%S')}.log"
            segfault_path.rename(segfault_backup_path)

            # submit to Sentry
            submit_sentry_attachment("Segfault detected", segfault_backup_path)
        except PermissionError:
            with suppress(Exception):
                submit_sentry_attachment("Segfault detected", segfault_path)
            logger.exception("Failed to backup segfault file.")
            return

    except Exception:
        logger.exception("Failed to submit segfault to Sentry.")



import numpy as np
from koyo.rand import get_random_seed, temporary_seed


def test_get_random_seed():
    rand = get_random_seed()
    assert rand > 0


def test_temporary_seed():
    with temporary_seed(0):
        values_first = np.random.rand(4)
    with temporary_seed(0):
        values_second = np.random.rand(4)
    assert np.all(values_first == values_second)



"""Various context managers."""

import gc
import os
import pathlib
import sys
from contextlib import contextmanager, redirect_stdout


@contextmanager
def fix_pathlib():
    """Fix pathlib on Windows."""
    posix_backup = pathlib.PosixPath
    try:
        pathlib.PosixPath = pathlib.WindowsPath
        yield
    finally:
        pathlib.PosixPath = posix_backup


@contextmanager
def hide_stdout():
    """Hide stdout."""
    with redirect_stdout(open(os.devnull, "w")):
        yield


@contextmanager
def nullout():
    """Context manager to suppress stdout and stderr."""
    save_stdout = sys.stdout
    save_stderr = sys.stderr
    sys.stdout = open(os.devnull, "w")
    sys.stderr = open(os.devnull, "w")
    try:
        yield
    finally:
        sys.stdout = save_stdout
        sys.stderr = save_stderr


@contextmanager
def noop(*args, **kwargs):
    """Function."""
    yield


@contextmanager
def no_gc():
    """Context manager to disable garbage collection."""
    enable = gc.isenabled()
    try:
        gc.disable()
        yield
    finally:
        if enable:
            gc.collect(0)
            gc.enable()


disabled_gc = no_gc



"""Dataframe utilities."""

from __future__ import annotations

import typing as ty
from pathlib import Path

from koyo.typing import PathLike

if ty.TYPE_CHECKING:
    import pandas as pd


def read_csv_with_comments(path: PathLike) -> pd.DataFrame:
    """Read CSV with comments."""
    import pandas as pd

    path = Path(path)
    try:
        df = pd.read_csv(path)
        first_col = df.iloc[:, 0]
        # check whether any of the values start with # (comment)
        for value in first_col:
            if value.startswith("# "):
                df = df[~first_col.str.contains("# ")]
                df.reset_index(drop=True, inplace=True)
                df.columns = df.iloc[0]
                df.drop(df.index[0], inplace=True)
                break
    except pd.errors.ParserError:
        from io import StringIO

        data = path.read_text().split("\n")
        start_index, end_index = 0, 0
        for row in data:
            if row.startswith("#"):
                start_index += 1
                continue
            elif not row:
                end_index += 1
        df = pd.read_csv(StringIO("\n".join(data[start_index:-end_index])), sep=",")
    except Exception:
        raise pd.errors.ParserError(f"Failed to parse grid '{path}'.")
    return df



"""Project interface.

In multiple projects in the Van de Plas lab we use the 'project' interface. It basically controls how we organize
our data and how we interact with it.
"""

from __future__ import annotations

import typing as ty
from contextlib import contextmanager
from pathlib import Path

from loguru import logger

from koyo.typing import PathLike

logger = logger.bind(src="Project")


def is_none(value: ty.Any, default: ty.Any) -> ty.Any:
    """Returns the value if value is not None, otherwise it returns the default."""
    return value if value is not None else default


class ProjectBase:
    """Project base."""

    BASE_FILENAME: str = "project.config.json"
    PROJECT_SUFFIX: str = ".project"

    # config attributes
    DATASETS_KEY: ty.Literal["datasets"] = "datasets"
    METADATA_KEY: ty.Literal["metadata"] = "metadata"
    PARAMS_KEY: ty.Literal["parameters"] = "parameters"
    EXECUTION_KEY: ty.Literal["execution"] = "execution"

    def __init__(self, project_dir: PathLike, cache_dir: PathLike | None = None):
        self.project_dir = Path(project_dir)
        if not self.project_dir.suffix == self.PROJECT_SUFFIX:
            self.project_dir = self.project_dir.with_suffix(self.PROJECT_SUFFIX)
        self.project_dir.mkdir(exist_ok=True, parents=True)
        self._cache_dir = cache_dir
        self._log_dir = self.project_dir / "Logs"
        self._figures_dir = self.project_dir / "Figures"
        self._results_dir = self.project_dir / "Results"
        self._config = None

    @property
    def name(self) -> str:
        """Return name of the project."""
        return self.project_dir.stem

    @property
    def filename(self) -> Path:
        """Get path to the project configuration file."""
        return self.project_dir / self.BASE_FILENAME

    @classmethod
    def new(cls, output_dir: PathLike, name: str = "Project", open_if_exists: bool = True) -> ProjectBase:
        """Create new project."""
        project_dir = (Path(output_dir) / name).with_suffix(".annotine")
        if project_dir.exists():
            if open_if_exists:
                return cls(project_dir)
            raise OSError("Project with this name already exists.")
        logger.info(f"Created new project '{project_dir.name}'")
        return cls(project_dir)

    @property
    def cache_dir(self) -> Path:
        """Return path to cache directory."""
        if self._cache_dir is None:
            self._cache_dir = self.project_dir / "Cache"
        self._cache_dir.mkdir(parents=True, exist_ok=True)  # type: ignore[union-attr]
        return self._cache_dir  # type: ignore[return-value]

    @cache_dir.setter
    def cache_dir(self, value: PathLike) -> None:
        self._cache_dir = Path(value)

    @property
    def log_dir(self) -> Path:
        """Return path to Logs directory."""
        if self._log_dir is None:
            self._log_dir = self.project_dir / "Logs"
        self._log_dir.mkdir(parents=True, exist_ok=True)
        return self._log_dir

    @property
    def figures_dir(self) -> Path:
        """Return path to Figure directory."""
        if self._figures_dir is None:
            self._figures_dir = self.project_dir / "Figures"
        self._figures_dir.mkdir(parents=True, exist_ok=True)
        return self._figures_dir

    @property
    def results_dir(self) -> Path:
        """Get path or dictionary of paths."""
        results_dir = self.project_dir / "Results"
        results_dir.mkdir(exist_ok=True, parents=True)
        return results_dir

    @property
    def project_config(self) -> dict:
        """Get dict containing project configuration."""
        from json import load
        from json.decoder import JSONDecodeError

        if self._config is None:
            if self.filename.exists():
                try:
                    with open(self.filename) as f_ptr:
                        data = load(f_ptr)
                    self._config = self._parse_config(data)
                except JSONDecodeError:
                    self._config = {}
            else:
                self._config = {}
            self._validate_config()
        return self._config

    def validate_inputs(self) -> None:
        """Validate input paths."""
        raise NotImplementedError("Must implement method")

    def _parse_config(self, config: dict) -> dict:
        """Parse configuration."""
        return config

    def _validate_config(self) -> None:
        """Validate config."""
        raise NotImplementedError("Must implement method")

    def _cleanup_config(self) -> dict:
        """Cleanup config."""
        raise NotImplementedError("Must implement method")

    def save(self) -> None:
        """Save project configuration."""
        self._export()
        logger.info(f"Project configuration saved to {self.filename}")

    @contextmanager
    def autosave(self) -> ty.Generator[None, None, None]:
        """Context manager to automatically save project configuration."""
        yield
        self._export()

    def _export(self) -> None:
        """Export configuration file."""
        from json import dump

        if not self.project_config or self.project_config is None:
            raise ValueError("Project configuration file is not setup.")

        # create backup of the project file
        if self.filename.exists():
            backup_filename = self.filename.with_suffix(".bak")
            backup_filename.write_text(self.filename.read_text())

        # validate config
        self._validate_config()
        project_config = self._cleanup_config()
        with open(self.filename, "w") as f_ptr:
            dump(project_config, f_ptr, indent=2)

    @property
    def n_datasets(self) -> int:
        """Get number of datasets."""
        return len(self.datasets)

    @property
    def datasets(self) -> list[str]:
        """List set of available datasets that have been registered to the project."""
        return list(self.project_config.get(self.DATASETS_KEY, {}).keys())

    @property
    def paths(self) -> list[Path]:
        """List set of available datasets that have been registered to the project."""
        return list(map(Path, self.project_config.get(self.DATASETS_KEY, {}).values()))

    def dataset_iter(self) -> ty.Iterable[str]:
        """Iterate dataset name and path."""
        yield from self.project_config.get(self.DATASETS_KEY, {}).keys()

    def dataset_path_iter(self) -> ty.Iterator[tuple[str, Path]]:
        """Iterator of dataset:path values."""
        yield from zip(self.datasets, self.paths)



"""Test that click produces the expected output."""

import pytest
from koyo.click import parse_arg


@pytest.mark.parametrize(
    "arg, expected",
    [
        ("n=5", ("n", 5)),
        ("n=5.0", ("n", 5.0)),
        ("n=[10,20,30]", ("n", [10, 20, 30])),
        ("n=value=with=equal=sign", ("n", "value=with=equal=sign")),
    ],
)
def test_parse_arg(arg, expected):
    res_key, res_value = parse_arg(arg, "")
    assert res_key == expected[0], f"Expected {expected[0]}, got {res_key}"
    assert res_value == expected[1], f"Expected {expected[1]}, got {res_value}"


@pytest.mark.parametrize(
    "arg, expected",
    [
        ("--un:n=5", ("n", 5)),
        ("--un:n=5.0", ("n", 5.0)),
        ("--un:n=[10,20,30]", ("n", [10, 20, 30])),
    ],
)
def test_parse_arg(arg, expected):
    res_key, res_value = parse_arg(arg, "--un:")
    assert res_key == expected[0], f"Expected {expected[0]}, got {res_key}"
    assert res_value == expected[1], f"Expected {expected[1]}, got {res_value}"
